# Author: Su Ye
# This script is an example for generating block-based stack files from original ARD zip as intermediate inputs to the
# COLD algorithm in a HPC environment. As preparation, you need to download '_BT' and '_SR' for all Landsat
# ARD collection 1.
# This script has 4 steps: 1) warp single-path array to limit the observation inputs for each pixel
# if single_path is set True; 2) unzip all images and unpack bit-based QA bands; 3) partition each 5000*5000 temporal
# images to blocks and eliminate those image blocks if no clear, water or snow pixel were in it (so to save disk space
# and  enable independent IO for individual block in later time-series analysis); 4) save each image block to
# python-native binary format (.npy) into its block folders

# For a 42-year Landsat ARD C1 tile (~3000 images), this script averagely produces ~350 G intermediate disk
# files, and takes ~12 mins to finish if 200 EPYC 7452 cores are used.
import os
import sys
import shutil
import tarfile
import logging
import time
import datetime as dt
import xml.etree.ElementTree as ET
import yaml
import click
from pytz import timezone
from datetime import datetime
from os import listdir
from os.path import isfile, join, isdir
from pathlib import Path
from importlib import resources as importlib_resources

import pandas as pd
import numpy as geek
import numpy as np
from glob import glob
from dateutil.parser import parse
from osgeo import gdal_array
from osgeo import gdal
import fiona


# define constant here
QA_CLEAR = 0
QA_WATER = 1
QA_SHADOW = 2
QA_SNOW = 3
QA_CLOUD = 4
QA_FILL = 255

QA_CIRRUS_HLS = 0
QA_WATER_HLS = 5
QA_SHADOW_HLS = 3
QA_SNOW_HLS = 4
QA_CLOUDADJACENT_HLS = 2
QA_CLOUD_HLS = 1
res = 30


s2_stack_bands = ['B02', 'B03', 'B04', 'B8A', 'B11', 'B12', 'Fmask']
l8_stack_bands = ['B02', 'B03', 'B04', 'B05', 'B06', 'B07', 'Fmask']


def mask_value(vector, val):
    """
    Build a boolean mask around a certain value in the vector.

    Args:
        vector: 1-d ndarray of values
        val: values to mask on
    Returns:
        1-d boolean ndarray
    """
    return vector == val


def qabitval_array_HLS(packedint_array):
    """
    Institute a hierarchy of qa values that may be flagged in the bitpacked
    value.

    fill > cloud > shadow > snow > water > clear

    Args:
        packedint: int value to bit check
    Returns:
        offset value to use
    """
    unpacked = np.full(packedint_array.shape, 0)
    QA_CLOUD_unpacked = geek.bitwise_and(packedint_array, 1 << 1)
    QA_CLOUD_ADJ = geek.bitwise_and(packedint_array, 1 << 2)
    QA_SHADOW_unpacked = geek.bitwise_and(packedint_array, 1 << 3)
    QA_SNOW_unpacked = geek.bitwise_and(packedint_array, 1 << 4)
    QA_WATER_unpacked = geek.bitwise_and(packedint_array, 1 << 5)

    unpacked[QA_WATER_unpacked > 0] = QA_WATER
    unpacked[QA_SNOW_unpacked > 0] = QA_SNOW
    unpacked[QA_SHADOW_unpacked > 0] = QA_SHADOW
    unpacked[QA_CLOUD_ADJ > 0] = QA_CLOUD
    unpacked[QA_CLOUD_unpacked > 0] = QA_CLOUD
    unpacked[packedint_array == QA_FILL] = QA_FILL

    return unpacked


def qabitval_array(packedint_array):
    """
    Institute a hierarchy of qa values that may be flagged in the bitpacked
    value.
    fill > cloud > shadow > snow > water > clear
    Args:
        packedint: int value to bit check
    Returns:
        offset value to use
    """
    unpacked = np.full(packedint_array.shape, QA_FILL)
    QA_CLOUD_unpacked = geek.bitwise_and(packedint_array, 1 << (QA_CLOUD + 1))
    QA_SHADOW_unpacked = geek.bitwise_and(packedint_array, 1 << (QA_SHADOW + 1))
    QA_SNOW_unpacked = geek.bitwise_and(packedint_array, 1 << (QA_SNOW + 1))
    QA_WATER_unpacked = geek.bitwise_and(packedint_array, 1 << (QA_WATER + 1))
    QA_CLEAR_unpacked = geek.bitwise_and(packedint_array, 1 << (QA_CLEAR + 1))

    unpacked[QA_CLEAR_unpacked > 0] = QA_CLEAR
    unpacked[QA_WATER_unpacked > 0] = QA_WATER
    unpacked[QA_SNOW_unpacked > 0] = QA_SNOW
    unpacked[QA_SHADOW_unpacked > 0] = QA_SHADOW
    unpacked[QA_CLOUD_unpacked > 0] = QA_CLOUD
    return unpacked


def qabitval_array_c2(packedint_array):
    """
    Institute a hierarchy of qa values that may be flagged in the bitpacked
    value for c2

    fill > cloud > shadow > snow > water > clear

    Args:
        packedint: int value to bit check
    Returns:
        offset value to use
    """
    unpacked = np.full(packedint_array.shape, QA_FILL)
    QA_CLEAR_unpacked = geek.bitwise_and(packedint_array, 1 << 6)
    QA_SHADOW_unpacked = geek.bitwise_and(packedint_array, 1 << 4)
    QA_CLOUD_unpacked = geek.bitwise_and(packedint_array, 1 << 3)
    QA_DILATED_unpacked = geek.bitwise_and(packedint_array, 1 << 1)
    QA_SNOW_unpacked = geek.bitwise_and(packedint_array, 1 << 5)
    QA_WATER_unpacked = geek.bitwise_and(packedint_array, 1 << 7)

    unpacked[QA_SNOW_unpacked > 0] = QA_SNOW
    unpacked[QA_SHADOW_unpacked > 0] = QA_SHADOW
    unpacked[QA_CLOUD_unpacked > 0] = QA_CLOUD
    unpacked[QA_DILATED_unpacked > 0] = QA_CLOUD
    unpacked[QA_CLEAR_unpacked > 0] = QA_CLEAR
    unpacked[QA_WATER_unpacked > 0] = QA_WATER
    return unpacked


def load_data(file_name, gdal_driver='GTiff'):
    '''
    Converts a GDAL compatable file into a numpy array and associated geodata.
    The rray is provided so you can run with your processing - the geodata consists of the geotransform and gdal dataset object
    if you're using an ENVI binary as input, this willr equire an associated .hdr file otherwise this will fail.
    This needs modifying if you're dealing with multiple bands.

    VARIABLES
    file_name : file name and path of your file

    RETURNS
    image array
    (geotransform, inDs)
    '''
    driver_t = gdal.GetDriverByName(gdal_driver)  # http://www.gdal.org/formats_list.html
    driver_t.Register()

    inDs = gdal.Open(file_name, gdal.GA_ReadOnly)
    # print(inDs)
    if inDs is None:
        print('Couldnt open this file {}'.format(file_name))
        sys.exit("Try again!")

    # Extract some info form the inDs
    geotransform = inDs.GetGeoTransform()

    # Get the data as a numpy array
    band = inDs.GetRasterBand(1)
    cols = inDs.RasterXSize
    rows = inDs.RasterYSize
    image_array = band.ReadAsArray(0, 0, cols, rows)

    return image_array, (geotransform, inDs)


def single_image_stacking_hls(source_dir, out_dir, logger, config, folder, is_partition=True, clear_threshold=0,
                              low_date_bound=None, upp_date_bound=None):
    """
    unzip single image, convert bit-pack qa to byte value, and save as numpy
    :param source_dir: the parent folder to save image 'folder'
    :param out_dir: the folder to save result
    :param folder: the folder name of image
    :param logger: the handler of logger file
    :param config
    :param is_partition: True, partition each image into blocks; False, save original size of image
    :param clear_threshold: threshold of clear pixel percentage, if lower than threshold, won't be processed
    :param low_date_bound: the lower date of user interested year range
    :param upp_date_bound: the upper date of user interested year range
    :return:
    """
    try:
        QA_band = gdal_array.LoadFile(join(join(source_dir, folder), "{}.Fmask.tif".format(folder)))
    except ValueError as e:
        # logger.error('Cannot open QA band for {}: {}'.format(folder, e))
        logger.error('Cannot open QA band for {}: {}'.format(folder, e))
        return False

    # convertQA = np.vectorize(qabitval)
    QA_band_unpacked = qabitval_array_HLS(QA_band).astype(np.short)
    if clear_threshold > 0:
        clear_ratio = np.sum(np.logical_or(QA_band_unpacked == QA_CLEAR,
                                           QA_band_unpacked == QA_WATER)) \
                      / np.sum(QA_band_unpacked != QA_FILL)
    else:
        clear_ratio = 1

    if clear_ratio > clear_threshold:
        [collection, sensor, tile_id, imagetime, version1, version2] = folder.rsplit('.')
        year = imagetime[0:4]
        doy = imagetime[4:7]
        file_name = sensor + tile_id + year + doy + collection + version1
        if low_date_bound is not None:
            if pd.Timestamp.toordinal((dt.datetime(year, 1, 1) + dt.timedelta(doy - 1)).strftime('%Y-%m-%d')) < \
                    low_date_bound:
                return True
        if upp_date_bound is not None:
            if pd.Timestamp.toordinal((dt.datetime(year, 1, 1) + dt.timedelta(doy - 1)).strftime('%Y-%m-%d')) > \
                    upp_date_bound:
                return True

        if sensor == 'L30':
            try:
                B1 = gdal_array.LoadFile(join(join(source_dir, folder),
                                         "{}.B02.tif".format(folder)))
                B2 = gdal_array.LoadFile(join(join(source_dir, folder),
                                         "{}.B03.tif".format(folder)))
                B3 = gdal_array.LoadFile(join(join(source_dir, folder),
                                         "{}.B04.tif".format(folder)))
                B4 = gdal_array.LoadFile(join(join(source_dir, folder),
                                         "{}.B05.tif".format(folder)))
                B5 = gdal_array.LoadFile(join(join(source_dir, folder),
                                         "{}.B06.tif".format(folder)))
                B6 = gdal_array.LoadFile(join(join(source_dir, folder),
                                         "{}.B07.tif".format(folder)))
                B7 = np.full(B6.shape, 0)  # assign zero

            except Exception as e:
                # logger.error('Cannot open spectral bands for {}: {}'.format(folder, e))
                logger.error('Cannot open Landsat bands for {}: {}'.format(folder, e))
                return False
        elif sensor == 'S30':
            try:
                B1 = gdal_array.LoadFile(join(join(source_dir, folder),
                                              "{}.B02.tif".format(folder)))
                B2 = gdal_array.LoadFile(join(join(source_dir, folder),
                                              "{}.B03.tif".format(folder)))
                B3 = gdal_array.LoadFile(join(join(source_dir, folder),
                                              "{}.B04.tif".format(folder)))
                B4 = gdal_array.LoadFile(join(join(source_dir, folder),
                                              "{}.B8A.tif".format(folder)))
                B5 = gdal_array.LoadFile(join(join(source_dir, folder),
                                              "{}.B11.tif".format(folder)))
                B6 = gdal_array.LoadFile(join(join(source_dir, folder),
                                              "{}.B12.tif".format(folder)))
                B7 = np.full(B6.shape, 0)

            except Exception as e:
                # logger.error('Cannot open spectral bands for {}: {}'.format(folder, e))
                logger.error('Cannot open Landsat bands for {}: {}'.format(folder, e))
                return False

        if (B1 is None) or (B2 is None) or (B3 is None) or (B4 is None) or (B5 is None) or (B6 is None):
            logger.error('Reading Landsat band fails for {}'.format(folder))
            return False

        if is_partition is True:
            b_width = int(config['n_cols'] / config['n_block_x'])  # width of a block
            b_height = int(config['n_rows'] / config['n_block_y'])
            bytesize = 2  # short16 = 2 * byte
            # source: https://towardsdatascience.com/efficiently-splitting-an-image-into-tiles-in-python-using-numpy-d1bf0dd7b6f7
            B1_blocks = np.lib.stride_tricks.as_strided(B1, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B2_blocks = np.lib.stride_tricks.as_strided(B2, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B3_blocks = np.lib.stride_tricks.as_strided(B3, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B4_blocks = np.lib.stride_tricks.as_strided(B4, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B5_blocks = np.lib.stride_tricks.as_strided(B5, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B6_blocks = np.lib.stride_tricks.as_strided(B6, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B7_blocks = np.lib.stride_tricks.as_strided(B7, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            QA_blocks = np.lib.stride_tricks.as_strided(QA_band_unpacked,
                                                        shape=(config['n_block_y'],
                                                               config['n_block_x'], b_height,
                                                               b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize,
                                                                 bytesize))
            for i in range(config['n_block_y']):
                for j in range(config['n_block_x']):
                    # check if no valid pixels in the chip, then eliminate
                    qa_unique = np.unique(QA_blocks[i][j])

                    # skip blocks are all cloud, shadow or filled values
                    # in DHTC, we also don't need to save pixel that has qa value of 'QA_CLOUD',
                    # 'QA_SHADOW', or FILLED value (255)
                    if QA_CLEAR not in qa_unique and \
                            QA_WATER not in qa_unique and \
                            QA_SNOW not in qa_unique:
                        continue

                    block_folder = 'block_x{}_y{}'.format(j + 1, i + 1)
                    np.save(join(join(out_dir, block_folder), file_name),
                            np.dstack([B1_blocks[i][j], B2_blocks[i][j], B3_blocks[i][j], B4_blocks[i][j],
                                       B5_blocks[i][j], B6_blocks[i][j], B7_blocks[i][j],
                                       QA_blocks[i][j]]).astype(np.int16))

        else:
            np.save(join(out_dir, file_name),
                    np.dstack([B1, B2, B3, B4, B5, B6, B7, QA_band_unpacked]).astype(np.int16))
        # scene_list.append(folder_name)
    else:
        # logger.info('Not enough clear observations for {}'.format(folder[0:len(folder) - 3]))
        logger.warn('Not enough clear observations for {}'.format(folder))

    return True


def single_image_stacking_hls14(out_dir, logger, config, folder, is_partition=True, clear_threshold=0,
                                low_date_bound=None, upp_date_bound=None):
    """
    unzip single image, convert bit-pack qa to byte value, and save as numpy
    :param source_dir: the parent folder to save image 'folder'
    :param out_dir: the folder to save result
    :param folder: the folder name of image
    :param logger: the handler of logger file
    :param config
    :param is_partition: True, partition each image into blocks; False, save original size of image
    :param clear_threshold: threshold of clear pixel percentage, if lower than threshold, won't be processed
    :param low_date_bound: the lower bound of user interested date range
    :param upp_date_bound: the upper bound of user interested date range
    :return:
    """
    try:
        hdf_ds = gdal.Open(folder, gdal.GA_ReadOnly).GetSubDatasets()
        qa_ds = gdal.Open(hdf_ds[-1][0])
        QA_band = qa_ds.ReadAsArray()
    except ValueError as e:
        # logger.error('Cannot open QA band for {}: {}'.format(folder, e))
        logger.error('Cannot open QA band for {}: {}'.format(folder, e))
        return False

    # convertQA = np.vectorize(qabitval)
    QA_band_unpacked = qabitval_array_HLS(QA_band).astype(np.short)
    if clear_threshold > 0:
        clear_ratio = np.sum(np.logical_or(QA_band_unpacked == QA_CLEAR,
                                           QA_band_unpacked == QA_WATER)) \
                      / np.sum(QA_band_unpacked != QA_FILL)
    else:
        clear_ratio = 1

    if clear_ratio > clear_threshold:
        [collection, sensor, tile_id, imagetime, version1, version2, ext] = folder.split('/')[-1].rsplit('.')
        year = imagetime[0:4]
        doy = imagetime[4:7]
        file_name = sensor + tile_id + year + doy + collection + version1
        if low_date_bound is not None:
            if pd.Timestamp.toordinal(parse((dt.datetime(int(year), 1, 1) +
                                             dt.timedelta(int(doy) - 1)).strftime('%Y-%m-%d'))) < low_date_bound:
                return True

        if upp_date_bound is not None:
            if pd.Timestamp.toordinal(parse((dt.datetime(int(year), 1, 1) +
                                             dt.timedelta(int(doy) - 1)).strftime('%Y-%m-%d'))) > upp_date_bound:
                return True

        if sensor == 'L30':
            try:
                B1 = gdal.Open(hdf_ds[1][0]).ReadAsArray()
                B2 = gdal.Open(hdf_ds[2][0]).ReadAsArray()
                B3 = gdal.Open(hdf_ds[3][0]).ReadAsArray()
                B4 = gdal.Open(hdf_ds[4][0]).ReadAsArray()
                B5 = gdal.Open(hdf_ds[5][0]).ReadAsArray()
                B6 = gdal.Open(hdf_ds[6][0]).ReadAsArray()
                B7 = np.full(B6.shape, 0)  # assign zero

            except Exception as e:
                # logger.error('Cannot open spectral bands for {}: {}'.format(folder, e))
                logger.error('Cannot open Landsat bands for {}: {}'.format(folder, e))
                return False
        elif sensor == 'S30':
            try:
                B1 = gdal.Open(hdf_ds[1][0]).ReadAsArray()
                B2 = gdal.Open(hdf_ds[2][0]).ReadAsArray()
                B3 = gdal.Open(hdf_ds[3][0]).ReadAsArray()
                B4 = gdal.Open(hdf_ds[8][0]).ReadAsArray()
                B5 = gdal.Open(hdf_ds[11][0]).ReadAsArray()
                B6 = gdal.Open(hdf_ds[12][0]).ReadAsArray()
                B7 = np.full(B6.shape, 0)

            except Exception as e:
                # logger.error('Cannot open spectral bands for {}: {}'.format(folder, e))
                logger.error('Cannot open Landsat bands for {}: {}'.format(folder, e))
                return False

        if (B1 is None) or (B2 is None) or (B3 is None) or (B4 is None) or (B5 is None) or (B6 is None):
            logger.error('Reading Landsat band fails for {}'.format(folder))
            return False

        if is_partition is True:
            b_width = int(config['n_cols'] / config['n_block_x'])  # width of a block
            b_height = int(config['n_rows'] / config['n_block_y'])
            bytesize = 2  # short16 = 2 * byte
            # source: https://towardsdatascience.com/efficiently-splitting-an-image-into-tiles-in-python-using-numpy-d1bf0dd7b6f7
            B1_blocks = np.lib.stride_tricks.as_strided(B1, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B2_blocks = np.lib.stride_tricks.as_strided(B2, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B3_blocks = np.lib.stride_tricks.as_strided(B3, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B4_blocks = np.lib.stride_tricks.as_strided(B4, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B5_blocks = np.lib.stride_tricks.as_strided(B5, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B6_blocks = np.lib.stride_tricks.as_strided(B6, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B7_blocks = np.lib.stride_tricks.as_strided(B7, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            QA_blocks = np.lib.stride_tricks.as_strided(QA_band_unpacked,
                                                        shape=(config['n_block_y'],
                                                               config['n_block_x'], b_height,
                                                               b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize,
                                                                 bytesize))
            for i in range(config['n_block_y']):
                for j in range(config['n_block_x']):
                    # check if no valid pixels in the chip, then eliminate
                    qa_unique = np.unique(QA_blocks[i][j])

                    # skip blocks are all cloud, shadow or filled values
                    # in DHTC, we also don't need to save pixel that has qa value of 'QA_CLOUD',
                    # 'QA_SHADOW', or FILLED value (255)
                    if QA_CLEAR not in qa_unique and \
                            QA_WATER not in qa_unique and \
                            QA_SNOW not in qa_unique:
                        continue

                    block_folder = 'block_x{}_y{}'.format(j + 1, i + 1)
                    np.save(join(join(out_dir, block_folder), file_name),
                            np.dstack([B1_blocks[i][j], B2_blocks[i][j], B3_blocks[i][j], B4_blocks[i][j],
                                       B5_blocks[i][j], B6_blocks[i][j], B7_blocks[i][j],
                                       QA_blocks[i][j]]).astype(np.int16))

        else:
            np.save(join(out_dir, file_name),
                    np.dstack([B1, B2, B3, B4, B5, B6, B7, QA_band_unpacked]).astype(np.int16))
        # scene_list.append(folder_name)
    else:
        # logger.info('Not enough clear observations for {}'.format(folder[0:len(folder) - 3]))
        logger.warn('Not enough clear observations for {}'.format(folder))

    return True


def single_image_stacking(tmp_path, source_dir, out_dir, folder, clear_threshold, path_array, logger, config,
                            is_partition=True, low_date_bound=None, upp_date_bound=None):
    """
    unzip single image, convert bit-pack qa to byte value, and save as numpy
    :param tmp_path: tmp folder to save unzip image
    :param source_dir: image folder save source zipped files
    :param out_dir: the folder to save result
    :param folder: the folder name of image
    :param clear_threshold: threshold of clear pixel percentage, if lower than threshold, won't be processed
    :param path_array: path array has the same dimension of inputted image, and the pixel value indicates
                      the path which the pixel belongs to; if path_array == none, we will use all path
    :param logger: the handler of logger file
    :param config
    :param is_partition: True, partition each image into blocks; False, save original size of image
    :param low_date_bound: the lower bound of user interested year range
    :param upp_date_bound: the upper bound of user interested year range
    :return:
    """
    # unzip SR
    if os.path.exists(join(tmp_path, folder)):
        shutil.rmtree(join(tmp_path, folder), ignore_errors=True)
    if os.path.exists(join(tmp_path, folder.replace("SR", "BT"))):
        shutil.rmtree(join(tmp_path, folder.replace("SR", "BT")), ignore_errors=True)

    try:
        with tarfile.open(join(source_dir, folder + '.tar')) as tar_ref:
            try:
                tar_ref.extractall(join(tmp_path, folder))
            except Exception:
                # logger.warning('Unzip fails for {}'.format(folder))
                logger.error('Unzip fails for {}'.format(folder))
                return
    except IOError as e:
        logger.error('Unzip fails for {}: {}'.format(folder, e))
        # return

    # unzip BT
    try:
        with tarfile.open(join(source_dir, folder.replace("SR", "BT") + '.tar')) as tar_ref:
            try:
                tar_ref.extractall(join(tmp_path, folder.replace("SR", "BT")))
            except Exception:
                # logger.warning('Unzip fails for {}'.format(folder.replace("SR", "BT")))
                logger.error('Unzip fails for {}'.format(folder.replace("SR", "BT")))
                return
    except IOError as e:
        logger.error('Unzip fails for {}: {}'.format(folder.replace("SR", "BT"), e))
        return

    if not isdir(join(tmp_path, folder.replace("SR", "BT"))):
        logger.error('Fail to locate BT folder for {}'.format(folder))
        return

    try:
        QA_band = gdal_array.LoadFile(join(join(tmp_path, folder),
                                                   "{}_PIXELQA.tif".format(folder[0:len(folder) - 3])))
    except ValueError as e:
        # logger.error('Cannot open QA band for {}: {}'.format(folder, e))
        logger.error('Cannot open QA band for {}: {}'.format(folder, e))
        return

    # convertQA = np.vectorize(qabitval)
    QA_band_unpacked = qabitval_array(QA_band).astype(np.short)
    if clear_threshold > 0:
        clear_ratio = np.sum(np.logical_or(QA_band_unpacked == QA_CLEAR,
                                           QA_band_unpacked == QA_WATER)) \
                      / np.sum(QA_band_unpacked != QA_FILL)
    else:
        clear_ratio = 1

    if clear_ratio > clear_threshold:
        if folder[3] == '5':
            sensor = 'LT5'
        elif folder[3] == '7':
            sensor = 'LE7'
        elif folder[3] == '8':
            sensor = 'LC8'
        elif folder[3] == '4':
            sensor = 'LT4'
        elif folder[3] == '9':
            sensor = 'LC9'
        else:
            logger.error('Sensor is not correctly formatted for the scene {}'.format(folder))

        col = folder[8:11]
        row = folder[11:14]
        year = folder[15:19]
        doy = datetime(int(year), int(folder[19:21]), int(folder[21:23])).strftime('%j')
        collection = "C{}".format(folder[35:36])
        version = folder[37:40]
        file_name = sensor + col + row + year + doy + collection + version
        if low_date_bound is not None:
            if pd.Timestamp.toordinal(parse((dt.datetime(int(year), 1, 1) +
                                             dt.timedelta(int(doy) - 1)).strftime('%Y-%m-%d'))) < low_date_bound:
                return
        if upp_date_bound is not None:
            if pd.Timestamp.toordinal(parse((dt.datetime(int(year), 1, 1) +
                                             dt.timedelta(int(doy) - 1)).strftime('%Y-%m-%d'))) > upp_date_bound:
                return

        if sensor == 'LT5' or sensor == 'LE7' or sensor == 'LT4':
            try:
                B1 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                         "{}B1.tif".format(folder)))
                B2 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                         "{}B2.tif".format(folder)))
                B3 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                         "{}B3.tif".format(folder)))
                B4 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                         "{}B4.tif".format(folder)))
                B5 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                         "{}B5.tif".format(folder)))
                B6 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                         "{}B7.tif".format(folder)))
                B7 = gdal_array.LoadFile(
                    join(join(tmp_path, "{}_BT".format(folder[0:len(folder) - 3])),
                         "{}_BTB6.tif".format(folder[0:len(folder) - 3])))
            except ValueError as e:
                # logger.error('Cannot open spectral bands for {}: {}'.format(folder, e))
                logger.error('Cannot open Landsat bands for {}: {}'.format(folder, e))
                return
        elif sensor == 'LC8' or 'LC9':
            try:
                B1 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                              "{}B2.tif".format(folder)))
                B2 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                              "{}B3.tif".format(folder)))
                B3 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                              "{}B4.tif".format(folder)))
                B4 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                              "{}B5.tif".format(folder)))
                B5 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                              "{}B6.tif".format(folder)))
                B6 = gdal_array.LoadFile(join(join(tmp_path, folder),
                                              "{}B7.tif".format(folder)))
                B7 = gdal_array.LoadFile(
                    join(join(tmp_path, "{}_BT".format(folder[0:len(folder) - 3])),
                         "{}_BTB10.tif".format(folder[0:len(folder) - 3])))
            except ValueError as e:
                # logger.error('Cannot open spectral bands for {}: {}'.format(folder, e))
                logger.error('Cannot open Landsat bands for {}: {}'.format(folder, e))
                return

        if (B1 is None) or (B2 is None) or (B3 is None) or (B4 is None) or (B5 is None) or (B6 is None) or \
                (B7 is None):
            logger.error('Reading Landsat band fails for {}'.format(folder))
            return

        # if path_array is not None, we will eliminate those observation that has different path with its assigned path
        if path_array is not None:  # meaning that single-path processing
            if not os.path.exists(join(join(tmp_path, folder), folder.replace("_SR", ".xml"))):
                logger.error('Cannot find xml file for {}'.format(join(join(tmp_path, folder), folder.replace("_SR", ".xml"))))
                return
            tree = ET.parse(join(join(tmp_path, folder), folder.replace("_SR", ".xml")))

            # get root element
            root = tree.getroot()
            elements = root.findall(
                './{https://landsat.usgs.gov/ard/v1}scene_metadata/{https://landsat.usgs.gov/'
                'ard/v1}global_metadata/{https://landsat.usgs.gov/ard/v1}wrs')
            if len(elements) == 0:
                logger.error('Parsing xml fails for {}'.format(folder))
                return
            pathid = int(elements[0].attrib['path'])

            # assign filled value to the pixels has different path id so won't be processed
            QA_band_unpacked[path_array != pathid] = QA_FILL

        if is_partition is True:
            b_width = int(config['n_cols'] / config['n_block_x'])  # width of a block
            b_height = int(config['n_rows'] / config['n_block_y'])
            bytesize = 2  # short16 = 2 * byte
            # source: https://towardsdatascience.com/efficiently-splitting-an-image-into-tiles-in-python-using-numpy-d1bf0dd7b6f7
            B1_blocks = np.lib.stride_tricks.as_strided(B1, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B2_blocks = np.lib.stride_tricks.as_strided(B2, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B3_blocks = np.lib.stride_tricks.as_strided(B3, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B4_blocks = np.lib.stride_tricks.as_strided(B4, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B5_blocks = np.lib.stride_tricks.as_strided(B5, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B6_blocks = np.lib.stride_tricks.as_strided(B6, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B7_blocks = np.lib.stride_tricks.as_strided(B7, shape=(config['n_block_y'],
                                                        config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            QA_blocks = np.lib.stride_tricks.as_strided(QA_band_unpacked,
                                                        shape=(config['n_block_y'],
                                                               config['n_block_x'], b_height,
                                                               b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize,
                                                                 bytesize))
            for i in range(config['n_block_y']):
                for j in range(config['n_block_x']):
                    # check if no valid pixels in the chip, then eliminate
                    qa_unique = np.unique(QA_blocks[i][j])

                    # skip blocks are all cloud, shadow or filled values
                    # in DHTC, we also don't need to save pixel that has qa value of 'QA_CLOUD',
                    # 'QA_SHADOW', or FILLED value (255)
                    if QA_CLEAR not in qa_unique and \
                            QA_WATER not in qa_unique and \
                            QA_SNOW not in qa_unique:
                        continue

                    block_folder = 'block_x{}_y{}'.format(j + 1, i + 1)
                    np.save(join(join(out_dir, block_folder), file_name),
                            np.dstack([B1_blocks[i][j], B2_blocks[i][j], B3_blocks[i][j], B4_blocks[i][j],
                                       B5_blocks[i][j], B6_blocks[i][j], B7_blocks[i][j],
                                       QA_blocks[i][j]]).astype(np.int16))

        else:
            np.save(join(out_dir, file_name), np.dstack([B1, B2, B3, B4, B5, B6, B7, QA_band_unpacked]).astype(np.int16))
        # scene_list.append(folder_name)
    else:
        # logger.info('Not enough clear observations for {}'.format(folder[0:len(folder) - 3]))
        logger.warn('Not enough clear observations for {}'.format(folder[0:len(folder) - 3]))

    # delete unzip folder
    shutil.rmtree(join(tmp_path, folder), ignore_errors=True)
    shutil.rmtree(join(tmp_path, folder.replace("SR", "BT")), ignore_errors=True)


def single_image_stacking_collection2(tmp_path, source_dir, out_dir, folder, clear_threshold, logger, config, bounds,
                                      is_partition=True, low_date_bound=None, upp_date_bound=None):
    """
    for collection 2
    :param tmp_path: tmp folder to save unzip image
    :param source_dir: image folder save source zipped files
    :param out_dir: the folder to save result
    :param folder: the folder name of image
    :param clear_threshold: threshold of clear pixel percentage, if lower than threshold, won't be processed
    :param logger: the handler of logger file
    :param config
    :param is_partition: True, partition each image into blocks; False, save original size of image
    :param low_date_bound: the lower bound of user interested date range
    :param upp_date_bound: the upper bound of user interested date range
    :param bounds
    :return:
    """
    # unzip SR
    if os.path.exists(join(tmp_path, folder)):
        shutil.rmtree(join(tmp_path, folder), ignore_errors=True)

    try:
        with tarfile.open(join(source_dir, folder + '.tar')) as tar_ref:
            try:
                tar_ref.extractall(join(tmp_path, folder))
            except Exception:
                # logger.warning('Unzip fails for {}'.format(folder))
                logger.error('Unzip fails for {}'.format(folder))
                return
    except IOError as e:
        logger.error('Unzip fails for {}: {}'.format(folder, e))
        # return

    try:
        QA_band = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                               "{}_QA_PIXEL.TIF".format(folder))),
                            outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                            dstNodata=1, srcNodata=1, outputType=gdal.GDT_UInt16).ReadAsArray()
    except ValueError as e:
        # logger.error('Cannot open QA band for {}: {}'.format(folder, e))
        logger.error('Cannot open QA band for {}: {}'.format(folder, e))
        return

    # convertQA = np.vectorize(qabitval)
    QA_band_unpacked = qabitval_array_c2(QA_band).astype(np.short)
    if clear_threshold > 0:
        clear_ratio = np.sum(np.logical_or(QA_band_unpacked == QA_CLEAR,
                                           QA_band_unpacked == QA_WATER)) \
                      / np.sum(QA_band_unpacked != QA_FILL)
    else:
        clear_ratio = 1

    if clear_ratio > clear_threshold:
        if folder[3] == '5':
            sensor = 'LT5'
        elif folder[3] == '7':
            sensor = 'LE7'
        elif folder[3] == '8':
            sensor = 'LC8'
        elif folder[3] == '9':
            sensor = 'LC9'
        elif folder[3] == '4':
            sensor = 'LT4'
        else:
            logger.error('Sensor is not correctly formatted for the scene {}'.format(folder))

        path = folder[10:13]
        row = folder[13:16]
        year = folder[17:21]
        doy = datetime(int(year), int(folder[21:23]), int(folder[23:25])).strftime('%j')
        collection = "C2"
        version = folder[len(folder) - 2:len(folder)]
        file_name = sensor + path + row + year + doy + collection + version
        if low_date_bound is not None:
            if pd.Timestamp.toordinal(parse((dt.datetime(int(year), 1, 1) +
                                             dt.timedelta(int(doy) - 1)).strftime('%Y-%m-%d'))) < low_date_bound:
                return True

        if upp_date_bound is not None:
            if pd.Timestamp.toordinal(parse((dt.datetime(int(year), 1, 1) +
                                             dt.timedelta(int(doy) - 1)).strftime('%Y-%m-%d'))) > upp_date_bound:
                return True

        if sensor == 'LT5' or sensor == 'LE7' or sensor == 'LT4':
            try:
                # B1 = gdal_array.LoadFile(join(join(tmp_path, folder),
                #                               "{}_SR_B1.TIF".format(folder)))
                B1 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B1.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
                B2 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B2.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0,
                               outputType=gdal.GDT_UInt16).ReadAsArray()
                B3 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B3.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0,
                               outputType=gdal.GDT_UInt16).ReadAsArray()
                B4 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B4.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0,
                               outputType=gdal.GDT_UInt16).ReadAsArray()
                B5 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B5.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()

                B6 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B7.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
                B7 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_ST_B6.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
            except ValueError as e:
                # logger.error('Cannot open spectral bands for {}: {}'.format(folder, e))
                logger.error('Cannot open Landsat bands for {}: {}'.format(folder, e))
                return
        elif sensor == 'LC8' or 'LC9':
            try:
                B1 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B2.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
                B2 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B3.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
                B3 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B4.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
                B4 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B5.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
                B5 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B6.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
                B6 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_SR_B7.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
                B7 = gdal.Warp(os.path.join(tmp_path, '_tmp_img'), gdal.Open(join(join(tmp_path, folder),
                                                                                  "{}_ST_B10.TIF".format(folder))),
                               outputBounds=[bounds[0], bounds[1], bounds[2], bounds[3]], xRes=res, yRes=res,
                               dstNodata=0, srcNodata=0, outputType=gdal.GDT_UInt16).ReadAsArray()
            except ValueError as e:
                # logger.error('Cannot open spectral bands for {}: {}'.format(folder, e))
                logger.error('Cannot open Landsat bands for {}: {}'.format(folder, e))
                return

        if (B1 is None) or (B2 is None) or (B3 is None) or (B4 is None) or (B5 is None) or (B6 is None) or \
                (B7 is None):
            logger.error('Reading Landsat band fails for {}'.format(folder))
            return

        # source: https://www.usgs.gov/faqs/how-do-i-use-scale-factor-landsat-level-2-science-products?qt-
        # news_science_products=0#qt-news_science_products recommended by yongquan
        B1 = (10000 * (B1 * 2.75e-05 - 0.2)).astype(np.int16)
        B2 = (10000 * (B2 * 2.75e-05 - 0.2)).astype(np.int16)
        B3 = (10000 * (B3 * 2.75e-05 - 0.2)).astype(np.int16)
        B4 = (10000 * (B4 * 2.75e-05 - 0.2)).astype(np.int16)
        B5 = (10000 * (B5 * 2.75e-05 - 0.2)).astype(np.int16)
        B7 = (10000 * (B7 * 2.75e-05 - 0.2)).astype(np.int16)
        B6 = (10 * (B6 * 0.00341802 + 149)).astype(np.int16)

        if is_partition is True:
            b_width = int(config['n_cols'] / config['n_block_x'])  # width of a block
            b_height = int(config['n_rows'] / config['n_block_y'])
            bytesize = 2  # short16 = 2 * byte
            # source: https://towardsdatascience.com/efficiently-splitting-an-image-into-tiles-
            # in-python-using-numpy-d1bf0dd7b6f7
            B1_blocks = np.lib.stride_tricks.as_strided(B1, shape=(config['n_block_y'],
                                                                   config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B2_blocks = np.lib.stride_tricks.as_strided(B2, shape=(config['n_block_y'],
                                                                   config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B3_blocks = np.lib.stride_tricks.as_strided(B3, shape=(config['n_block_y'],
                                                                   config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B4_blocks = np.lib.stride_tricks.as_strided(B4, shape=(config['n_block_y'],
                                                                   config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B5_blocks = np.lib.stride_tricks.as_strided(B5, shape=(config['n_block_y'],
                                                                   config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B6_blocks = np.lib.stride_tricks.as_strided(B6, shape=(config['n_block_y'],
                                                                   config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            B7_blocks = np.lib.stride_tricks.as_strided(B7, shape=(config['n_block_y'],
                                                                   config['n_block_x'], b_height, b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize, bytesize))
            QA_blocks = np.lib.stride_tricks.as_strided(QA_band_unpacked,
                                                        shape=(config['n_block_y'],
                                                               config['n_block_x'], b_height,
                                                               b_width),
                                                        strides=(config['n_cols'] * b_height * bytesize,
                                                                 b_width * bytesize,
                                                                 config['n_cols'] * bytesize,
                                                                 bytesize))
            for i in range(config['n_block_y']):
                for j in range(config['n_block_x']):
                    # check if no valid pixels in the chip, then eliminate
                    qa_unique = np.unique(QA_blocks[i][j])

                    # skip blocks are all cloud, shadow or filled values
                    # in DHTC, we also don't need to save pixel that has qa value of 'QA_CLOUD',
                    # 'QA_SHADOW', or FILLED value (255)
                    if QA_CLEAR not in qa_unique and \
                            QA_WATER not in qa_unique and \
                            QA_SNOW not in qa_unique:
                        continue

                    block_folder = 'block_x{}_y{}'.format(j + 1, i + 1)
                    np.save(join(join(out_dir, block_folder), file_name),
                            np.dstack([B1_blocks[i][j], B2_blocks[i][j], B3_blocks[i][j], B4_blocks[i][j],
                                       B5_blocks[i][j], B6_blocks[i][j], B7_blocks[i][j], QA_blocks[i][j]]).astype(np.int16))

        else:
            np.save(join(out_dir, file_name), np.dstack([B1, B2, B3, B4, B5, B6, B7, QA_band_unpacked]).astype(np.int16))
        # scene_list.append(folder_name)
    else:
        # logger.info('Not enough clear observations for {}'.format(folder[0:len(folder) - 3]))
        logger.warn('Not enough clear observations for {}'.format(folder[0:len(folder) - 3]))

    # delete unzip folder
    shutil.rmtree(join(tmp_path, folder), ignore_errors=True)
    shutil.rmtree(join(tmp_path, folder.replace("SR", "BT")), ignore_errors=True)


def checkfinished_step1(tmp_path):
    """
    :param tmp_path:
    :return:
    """
    if not os.path.exists(tmp_path):
        return False
    return True


def checkfinished_step2(out_dir, n_cores):
    """
    :param out_dir:
    :param n_cores:
    :return:
    """
    for i in range(n_cores):
        if not os.path.exists(join(out_dir, 'rank{}_finished.txt'.format(i + 1))):
            return False
    return True


def checkfinished_step3_partition(out_dir):
    """
    :param out_dir:
    :return:
    """
    if not os.path.exists(join(out_dir, "starting_last_dates.txt")):
        return False
    else:
        return True


def checkfinished_step3_nopartition(out_dir):
    """
    :param out_dir:
    :return:
    """
    if not os.path.exists(join(out_dir, "scene_list.txt")):
        return False
    return True


def get_extent(extent_geojson, res, buf=0):
    """
    read shapefile of a tile from an S3 bucket, and convert projection to be aligned with sample image.
    arg:
        'extent_geojson': sharply geojson object
        res: planet resolution
    return:
        (float, float, float, float), (int, int)) tuple
    """
    # txmin = min([row[0] for row in extent_geojson['coordinates'][0]]) - res / 2.0
    # txmax = max([row[0] for row in extent_geojson['coordinates'][0]]) + res / 2.0
    # tymin = min([row[1] for row in extent_geojson['coordinates'][0]]) - res / 2.0
    # tymax = max([row[1] for row in extent_geojson['coordinates'][0]]) + res / 2.0
    txmin = extent_geojson['bbox'][0] - res * buf
    txmax = extent_geojson['bbox'][2] + res * buf
    tymin = extent_geojson['bbox'][1] - res * buf
    tymax = extent_geojson['bbox'][3] + res * buf
    n_row = np.ceil((tymax - tymin) / res)
    n_col = np.ceil((txmax - txmin) / res)
    txmin_new = (txmin + txmax) / 2 - n_col / 2 * res
    txmax_new = (txmin + txmax) / 2 + n_col / 2 * res
    tymin_new = (tymin + tymax) / 2 - n_row / 2 * res
    tymax_new = (tymin + tymax) / 2 + n_row / 2 * res
    return (txmin_new, txmax_new, tymin_new, tymax_new), (n_row, n_col)


def get_feature(shp, id):
    for feature in shp:
        if feature['properties']['id'] == id:
            return feature


def explode(coords):
    """Explode a GeoJSON geometry's coordinates object and yield coordinate tuples.
    As long as the input is conforming, the type of the geometry doesn't matter."""
    for e in coords:
        if isinstance(e, (float, int)):
            yield coords
            break
        else:
            for f in explode(e):
                yield f


def bbox(f):
    x, y = zip(*list(explode(f['geometry']['coordinates'])))
    return min(x), min(y), max(x), max(y)


@click.command()
@click.option('--source_dir', type=str, default=None, help='the folder directory of Landsat tar files '
                                                           'downloaded from USGS website')
@click.option('--out_dir', type=str, default=None, help='the folder directory for storing stacks')
@click.option('--clear_threshold', type=float, default=0, help='user-defined clear pixel proportion')
@click.option('--single_path', type=bool, default=True, help='indicate if using single_path or sidelap')
@click.option('--rank', type=int, default=1, help='the rank id')
@click.option('--n_cores', type=int, default=1, help='the total cores assigned')
@click.option('--is_partition/--continuous', default=True, help='partition the output to blocks')
@click.option('--yaml_path', type=str, help='yaml file path')
@click.option('--hpc/--dhtc', default=True, help='if it is set for HPC or DHTC environment')
@click.option('--low_date_bound', type=str, default=None, help='the lower bound of the year range of user interest')
@click.option('--upp_date_bound', type=str, default=None, help='the upper bound of the year range of user interest')
@click.option('--collection', type=click.Choice(['ARD', 'C2', 'HLS', 'HLS14']), default='ARD',
              help='image source')
@click.option('--shapefile_path', type=str, default=None)
@click.option('--id', type=int, default=0)
def main(source_dir, out_dir, clear_threshold, single_path, rank, n_cores, is_partition, yaml_path, hpc, low_date_bound,
         upp_date_bound, collection, shapefile_path, id):
    if not os.path.exists(source_dir):
        print('Source directory not exists!')

    # select only _SR
    if collection == 'ARD':
        folder_list = [f[0:len(f) - 4] for f in listdir(source_dir) if
                       (isfile(join(source_dir, f)) and f.endswith('.tar')
                        and f[len(f) - 6:len(f) - 4] == 'SR')]
    elif collection == 'C2':
        folder_list = [f[0:len(f) - 4] for f in listdir(source_dir) if
                       (isfile(join(source_dir, f)) and f.endswith('.tar'))]
    elif collection == 'HLS':
        folder_list = [f for f in listdir(source_dir) if f.startswith('HLS')]
    elif collection == 'HLS14':
        folder_list = [y for x in os.walk(source_dir) for y in glob(os.path.join(x[0], '*.hdf'))]

    tmp_path = join(out_dir, 'tmp{}'.format(rank))

    tz = timezone('US/Eastern')
    with open(yaml_path) as yaml_obj:
        config = yaml.safe_load(yaml_obj)

    logger = None
    # create needed folders
    if rank == 1:
        logging.basicConfig(filename=join(os.getcwd(), 'prepare_ard.log'),
                            filemode='w', level=logging.INFO)  # mode = w enables the log file to be overwritten
        logger = logging.getLogger(__name__)

        logger.info('AutoPrepareDataARD starts: {}'.format(datetime.now(tz).strftime('%Y-%m-%d %H:%M:%S')))

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        if is_partition is True:
            for i in range(config['n_block_y']):
                for j in range(config['n_block_x']):
                    block_folder = 'block_x{}_y{}'.format(j + 1, i + 1)
                    if not os.path.exists(join(out_dir, block_folder)):
                        os.mkdir(join(out_dir, block_folder))

        if not os.path.exists(tmp_path):
            os.mkdir(tmp_path)

        if hpc is True and collection == 'ARD':
            # warp a tile-based single path tif

            with importlib_resources.path('pycold.imagetool', 'singlepath_landsat_conus.tif') as conus_image_fpath:
                # conus_image_fpath = (Path(__file__).parent / 'singlepath_landsat_conus.tif').resolve()
                conus_image = gdal.Open(os.fspath(conus_image_fpath))

            if os.path.exists(join(tmp_path, folder_list[0])):
                shutil.rmtree(join(tmp_path, folder_list[0]), ignore_errors=True)
            with tarfile.open(join(source_dir, folder_list[0] + '.tar')) as tar_ref:
                try:
                    tar_ref.extractall(join(tmp_path, folder_list[0]))
                except Exception:
                    logger.error('Unzip fails for {}'.format(folder_list[0]))
            ref_image = gdal.Open(join(join(tmp_path, folder_list[0]), "{}B1.tif".format(folder_list[0])))
            trans = ref_image.GetGeoTransform()
            proj = ref_image.GetProjection()
            xmin = trans[0]
            ymax = trans[3]
            xmax = xmin + trans[1] * ref_image.RasterXSize
            ymin = ymax + trans[5] * ref_image.RasterYSize
            params = gdal.WarpOptions(dstSRS=proj, outputBounds=[xmin, ymin, xmax, ymax],
                                      width=ref_image.RasterXSize, height=ref_image.RasterYSize)

            out_fpath = join(out_dir, 'singlepath_landsat_tile.tif')
            dst = gdal.Warp(out_fpath, conus_image, options=params)
            # must close the dst
            dst = None  # NOQA
            out_img = None  # NOQA
            shutil.rmtree(join(tmp_path, folder_list[0]), ignore_errors=True)

        if collection == 'ARD':
            ordinal_dates = [pd.Timestamp.toordinal(dt.datetime(int(folder[15:19]), int(folder[19:21]),
                                                                int(folder[21:23])))
                             for folder in folder_list]
        elif collection == 'C2':
            ordinal_dates = [pd.Timestamp.toordinal(dt.datetime(int(folder[17:21]), int(folder[21:23]),
                                                                int(folder[23:25])))
                             for folder in folder_list]
        elif collection == 'HLS':
            ordinal_dates = [pd.Timestamp.toordinal(dt.datetime(int(folder[15:19]), 1, 1)) + int(folder[19:22]) - 1
                             for folder in folder_list]
        elif collection == 'HLS14':
            ordinal_dates = [pd.Timestamp.toordinal(dt.datetime(int(folder.split('/')[-1][15:19]), 1, 1))
                             + int(folder.split('/')[-1][19:22]) - 1 for folder in folder_list]
        ordinal_dates.sort()
        file = open(join(out_dir, "starting_last_dates.txt"), "w+")  # need to save out starting and
        # lasting date for this tile
        file.writelines("{}\n".format(str(np.max([ordinal_dates[0],
                                                 pd.Timestamp.toordinal(dt.datetime(parse(low_date_bound)))]))))
        file.writelines("{}\n".format(str(np.min([ordinal_dates[-1],
                                                 pd.Timestamp.toordinal(dt.datetime(parse(upp_date_bound)))]))))
        file.close()
    else:
        logging.basicConfig(filename=join(os.getcwd(), 'prepare_ard.log'),
                            filemode='a', level=logging.INFO)  # mode = w enables the log file to be overwritten
        logger = logging.getLogger(__name__)

    # use starting_last_dates.txt to indicate all folder have been created. Very stupid way. Need improve it in future.
    while not checkfinished_step1(join(out_dir, "starting_last_dates.txt")):
        time.sleep(5)

    # read a general path file which can indicate which pixel is assigned to which path
    path_array = None
    if single_path is True and collection == 'ARD':
        # parse tile h and v from folder name
        folder_name = os.path.basename(source_dir)
        tile_h = int(folder_name[folder_name.find('h') + 1: folder_name.find('h') + 4])
        tile_v = int(folder_name[folder_name.find('v') + 1: folder_name.find('v') + 4])
        if hpc is True:
            path_array = gdal_array.LoadFile(join(out_dir, 'singlepath_landsat_tile.tif'))
        else:
            try:
                path_array = gdal_array.LoadFile(join(Path(os.path.realpath(__file__)).parent,
                                                      'singlepath_landsat_tile_crop_compress.tif'),
                                                 xoff=tile_h * config['n_cols'],
                                                 yoff=tile_v * config['n_rows'],
                                                 xsize=config['n_cols'],
                                                 ysize=config['n_rows'])  # read a partial array
                # from a large file
            except IOError as e:
                logger.error(e)
                exit()

    if collection == 'C2':
        # gpd_tile = gpd.read_file(shapefile_path)
        bounds = bbox(get_feature(fiona.open(shapefile_path), id))

        for i in range(int(np.ceil(len(folder_list) / n_cores))):
            new_rank = rank - 1 + i * n_cores
            if new_rank > (len(folder_list) - 1):  # means that all folder has been processed
                break
            folder = folder_list[new_rank]
            single_image_stacking_collection2(tmp_path, source_dir, out_dir, folder, clear_threshold, logger,
                                              config, bounds, is_partition=is_partition,
                                              low_date_bound=low_date_bound, upp_date_bound=upp_date_bound)
    elif collection == 'ARD':
        # assign files to each core
        for i in range(int(np.ceil(len(folder_list) / n_cores))):
            new_rank = rank - 1 + i * n_cores
            if new_rank > (len(folder_list) - 1):  # means that all folder has been processed
                break
            folder = folder_list[new_rank]
            single_image_stacking(tmp_path, source_dir, out_dir, folder, clear_threshold, path_array, logger,
                                  config, is_partition=is_partition, low_date_bound=low_date_bound,
                                  upp_date_bound=upp_date_bound)
    elif collection == 'HLS':
        # assign files to each core
        for i in range(int(np.ceil(len(folder_list) / n_cores))):
            new_rank = rank - 1 + i * n_cores
            if new_rank > (len(folder_list) - 1):  # means that all folder has been processed
                break
            folder = folder_list[new_rank]
            single_image_stacking_hls(source_dir, out_dir, logger, config, folder, clear_threshold=clear_threshold,
                                      is_partition=is_partition, low_date_bound=low_date_bound,
                                      upp_date_bound=upp_date_bound)
    elif collection == 'HLS14':
        # assign files to each core
        for i in range(int(np.ceil(len(folder_list) / n_cores))):
            new_rank = rank - 1 + i * n_cores
            if new_rank > (len(folder_list) - 1):  # means that all folder has been processed
                break
            folder = folder_list[new_rank]
            single_image_stacking_hls14(out_dir, logger, config, folder, clear_threshold=clear_threshold,
                                        is_partition=is_partition, low_date_bound=low_date_bound,
                                        upp_date_bound=upp_date_bound)
    # create an empty file for signaling the core that has been finished
    with open(os.path.join(out_dir, 'rank{}_finished.txt'.format(rank)), 'w'):
        pass

    # wait for other cores assigned
    # while not checkfinished_step2(out_dir, n_cores):
    #     time.sleep(5)
    shutil.rmtree(tmp_path, ignore_errors=True)
    # create scene list after stacking is finished and remove folders. Not need it in DHTC
    if rank == 1:
        # remove tmp folder
        logger.info('Stacking procedure finished: {}'.format(datetime.now(tz).strftime('%Y-%m-%d %H:%M:%S')))


if __name__ == '__main__':
    main()



# Autogenerated via:
# python ~/code/watch/dev/maintain/mirror_package_geowatch.py
def __getattr__(key):
    import watch.tasks.rutgers_material_change_detection.models.patch_transformer as mirror
    return getattr(mirror, key)


"""

python -m watch mlops "list"
python -m watch mlops "status"
python -m watch mlops "push pull evals"
python -m watch mlops "pull evals"
python -m watch mlops "pull packages"
python -m watch mlops "push evals"

SeeAlso:
    ~/code/watch/dev/reports/report_2022_09_xx.py

"""
import ubelt as ub
import math
import numpy as np
import pandas as pd
import functools  # NOQA
# APPLY Monkey Patches
# from watch import monkey
from watch import heuristics
from watch.mlops import smart_result_parser as smart


fit_param_keys = heuristics.fit_param_keys
pred_param_keys = heuristics.pred_param_keys
DSET_CODE_TO_GSD = heuristics.DSET_CODE_TO_GSD


def evaluation_report():
    """
    MAIN FUNCTION

    from watch.mlops.old.expt_report import *  # NOQA

    row = reporter.orig_merged_df.loc[121]
    print(ub.urepr(row.to_dict(), nl=1))
    path = reporter.orig_merged_df.loc[121]['raw']

    import platform
    host = platform.node()
    path.shrinkuser(home=f'$HOME/remote/{host}')


    print(ub.urepr(row.to_dict(), nl=1))
    """
    import kwplot
    kwplot.autosns()
    from watch import heuristics
    from watch.mlops import expt_manager
    expt_dvc_dpath = heuristics.auto_expt_dvc()
    manager = expt_manager.DVCExptManager.coerce(expt_dvc_dpath)
    state = manager
    reporter = EvaluationReporter(state)
    reporter.load()
    reporter.status()
    reporter.plot()

    if 0:
        merged_df = reporter.orig_merged_df.copy()
        merged_df[merged_df.expt.str.contains('invar')]['sc_macro_f1']
        merged_df[merged_df.in_production]['sc_macro_f1']

        selected = merged_df[merged_df.in_production].sort_values('sc_macro_f1')
        selected = selected[['siteprep_f1', 'active_f1', 'sc_macro_f1', 'model']]
        selected['sc_macro_f1'] = selected[['siteprep_f1', 'active_f1']].mean(axis=1)
        selected = selected.sort_values('sc_macro_f1')
        print(selected)


class EvaluationReporter:
    """
    Manages handing the data off to experiment plotting functions.
    """

    def __init__(reporter, state):
        """
        Args:
            state (DVCExptManager | ExperimentState)
        """
        reporter.state = state
        reporter.expt_dvc_dpath = state.expt_dvc_dpath

        reporter.raw_df = None
        reporter.filt_df = None
        reporter.comp_df = None

        reporter.dpath = ub.Path.appdir('watch/expt-report').ensuredir()

        reporter.metric_registry = pd.DataFrame([
            {'name': 'coi_mAP', 'tasks': ['sc'], 'human': 'Pixelwise mAP (classes of interest)'},
            {'name': 'coi_mAUC', 'tasks': ['sc'], 'human': 'Pixelwise mAUC (classes of interest)'},
            {'name': 'sc_macro_f1', 'tasks': ['sc'], 'human': 'IARPA SC mean F1'},

            {'name': 'salient_AP', 'tasks': ['bas'], 'human': 'Pixelwise Salient AP'},
            {'name': 'salient_AUC', 'tasks': ['bas'], 'human': 'Pixelwise Salient AUC'},
            {'name': 'BAS_F1', 'tasks': ['bas'], 'human': 'IARPA BAS F1'},
        ])
        reporter.metric_registry['type'] = 'metric'

        # TODO: add column types
        column_meanings = get_column_meanings()
        reporter.column_meanings = column_meanings

    def status(reporter, table=None):
        reporter.state.summarize()
        reporter.report_best(verbose=1)
        # if 0:
        #     if table is None:
        #         table = reporter.state.evaluation_table()
        #     loaded_table = load_extended_data(table, reporter.expt_dvc_dpath)
        #     loaded_table = pd.DataFrame(loaded_table)
        #     # dataset_summary_tables(dpath)
        #     initial_summary(table, loaded_table, reporter.dpath)

    def report_best(reporter, show_configs=0, verbose=0, top_k=2):
        """
        Ignore:
            show_configs = 0
            verbose = 1
            top_k = 2
        """
        import rich
        orig_merged_df = reporter.orig_merged_df
        metric_names = reporter.metric_registry.name

        cfg_names = reporter.state.hashed_cfgkeys
        id_names = ['trk_model', 'act_model'] + cfg_names

        # metric_names = reporter.metric_registry.name
        metric_names = [c for c in orig_merged_df.columns if 'metrics.' in c]
        resource_names = [c for c in orig_merged_df.columns if 'resource.' in c]
        metric_cols = (ub.oset(metric_names) & orig_merged_df.columns)
        resource_cols = (ub.oset(resource_names) & orig_merged_df.columns)

        metric_cols = [c for c in metric_cols if '.tau' not in c and '.rho' not in c]
        metric_cols = [c for c in metric_cols if '.salient_AUC' not in c]
        metric_cols = [c for c in metric_cols if '.salient_APUC' not in c]
        metric_cols = [c for c in metric_cols if '.macro_f1_active' not in c]
        metric_cols = [c for c in metric_cols if '.macro_f1_siteprep' not in c]
        metric_cols = [c for c in metric_cols if '.sc_micro_f1' not in c]
        metric_cols = [c for c in metric_cols if '.bas_ppv' not in c]
        metric_cols = [c for c in metric_cols if '.bas_tpr' not in c]
        metric_cols = [c for c in metric_cols if '.mAPUC' not in c]
        metric_cols = [c for c in metric_cols if '.mAUC' not in c]
        metric_cols = [c for c in metric_cols if ' ' not in c]
        # resource_cols = [c for c in resource_cols if 'total' in c or 'kwh' in c]
        resource_cols = [c for c in resource_cols if '.hardware' not in c]
        resource_cols = list(resource_cols)

        primary_metrics = (ub.oset(['act.poly.metrics.sc_macro_f1', 'trk.poly.metrics.bas_faa_f1']) & metric_cols)
        metric_cols = list((metric_cols & primary_metrics) | (metric_cols - primary_metrics))

        # print('orig_merged_df.columns = {}'.format(ub.urepr(list(orig_merged_df.columns), nl=1)))
        id_cols = list(ub.oset(id_names) & orig_merged_df.columns)

        # test_datasets = orig_merged_df['test_dset'].dropna().unique().tolist()
        # if verbose:
        #     rich.print('[orange1]-- REPORTING BEST --')
        #     print('test_datasets = {}'.format(ub.urepr(test_datasets, nl=1)))

        grouped_shortlists = {}
        group_keys = ub.oset(['test_trk_dset', 'test_act_dset', 'type'])
        group_keys = list(group_keys & orig_merged_df.columns.intersection(group_keys))

        def _condense_report_df(df):
            col_mapper = {c: c for c in df.columns}
            _metric_cols = []
            for k in col_mapper.keys():
                k2 = k.replace('metrics.', '')
                k2 = k2.replace('resource.', '')
                col_mapper[k] = k2
                _metric_cols.append(k2)
            df = df.rename(col_mapper, axis=1)
            drop_cols = [k for k in _metric_cols if df[k].isnull().all()]
            df = df.drop(drop_cols, axis=1)

            cpu_cols = [c for c in df.columns if '.cpu_name' in c]
            gpu_cols = [c for c in df.columns if '.gpu_name' in c]
            for c in cpu_cols:
                df[c] = df[c].apply(lambda x: x if not isinstance(x, str) else x.replace('Intel(R) Core(TM) ', ''))
            for c in gpu_cols:
                df[c] = df[c].apply(lambda x: x if not isinstance(x, str) else x.replace('NVIDIA GeForce ', ''))

            def collapse_redundant_columns(df, check_cols, new_name='cpu_name'):
                check_cols = df.columns.intersection(check_cols)
                if len(check_cols) > 0:
                    collapse_cpu_col = df[check_cols].apply(ub.allsame, axis=1).all()
                    if collapse_cpu_col:
                        df[new_name] = df[check_cols[0]]
                        df = df.drop(check_cols, axis=1)
                return df
            df = collapse_redundant_columns(df, cpu_cols, 'cpu')
            df = collapse_redundant_columns(df, gpu_cols, 'gpu')
            df = collapse_redundant_columns(df, ['trk.pxl.disk_type', 'act.pxl.disk_type'], 'disk_type')
            return df

        for groupid, subdf in orig_merged_df.groupby(group_keys):
            # group_row = ub.dzip(group_keys, groupid)
            # if '_poly_' not in group_row['type']:
            #     continue
            if verbose:
                print('')
                rich.print(f'[orange1] -- <BEST ON: {groupid}> --')

            top_indexes = set()
            for metric in metric_cols:
                best_rows = subdf[metric].sort_values()
                top_indexes.update(best_rows.iloc[-top_k:].index)

            idxs = sorted(top_indexes)
            shortlist = subdf.loc[idxs]
            shortlist = shortlist.sort_values(metric_cols, na_position='first')

            show_configs = show_configs

            if show_configs:
                cfg_cols = shortlist[ub.oset(cfg_names) & shortlist.columns]
                keys = [
                    v for v in ub.unique(cfg_cols.values[::-1].ravel())
                    if not pd.isnull(v)][::-1]
                resolved = reporter._build_cfg_rlut(keys)
                resolved = ub.udict(resolved).subdict(keys)
                if verbose and show_configs:
                    rich.print('resolved = {}'.format(ub.urepr(resolved, sort=0, nl=2)))

            # test_dset_to_best[test_dset] =
            if verbose:
                shortlist_small = shortlist[id_cols + metric_cols + resource_cols]
                shortlist_small = _condense_report_df(shortlist_small)
                # rich.print(shortlist_small.T.to_string())
                rich.print(shortlist_small.to_string())
                print('')
                rich.print(f'[orange1] -- </BEST ON: {groupid}> --')
                print('')
            grouped_shortlists[groupid] = shortlist

        return grouped_shortlists

    def serialize_rows(reporter):
        from kwcoco.util.util_json import ensure_json_serializable
        table = reporter.orig_merged_df
        # state = reporter.state

        colnames = ub.oset(reporter.orig_merged_df.columns)
        # column_nestings = util_dotdict.dotkeys_to_nested(colnames)
        dotted = ub.oset([c for c in colnames if '.' in c])
        metric_cols = ub.oset([c for c in dotted if 'metrics.' in c])
        meta_cols = ub.oset([c for c in dotted if 'meta.' in c])
        resource_cols = ub.oset([c for c in dotted if 'resource.' in c])
        fit_cols = ub.oset([c for c in dotted if 'fit.' in c])
        param_cols = dotted - (metric_cols | fit_cols | resource_cols | meta_cols)

        for row in table.to_dict('records'):
            row = ub.udict(ensure_json_serializable(row))
            walker = ub.IndexableWalker(row)
            for path, val in walker:
                if hasattr(val, 'spec'):
                    walker[path] = val.spec

            summary = {
                # 'model': row['model'],
                # 'file_name': pkg_fpath,
                'pred_params': row & param_cols,
                'fit_params': row & fit_cols,
                'resources': row & resource_cols,
                'meta': row & meta_cols,
                'metrics': row & metric_cols,
            }
            yield summary, row

    def build_analysis(reporter):
        from watch.utils.result_analysis import ResultAnalysis
        from watch.utils.result_analysis import Result
        # import json

        results = []
        for summary, details in reporter.serialize_rows():
            params = {
                **summary['fit_params'],
                **summary['pred_params'],
            }
            for k, v in params.items():
                if isinstance(v, list):
                    params[k] = repr(v)
            name = 'row_' + ub.hash_data(summary)[0:8]
            metrics = ub.udict(summary['metrics'])  # - {'properties'}
            result = Result(name, params, metrics)
            results.append(result)

        analysis = ResultAnalysis(
            results,
            metric_objectives={
                'act.poly.metrics.bas_f1': 'max',
                'act.poly.metrics.sc_macro_f1': 'max'
            },
            metrics=[
                # 'act.poly.metrics.bas_f1',
                'act.poly.metrics.sc_macro_f1'
            ]
        )
        analysis.build()
        analysis.report()
        return analysis

    def _build_cfg_rlut(reporter, keys):
        from watch.utils.reverse_hashid import ReverseHashTable
        if keys is None:
            candidates = ReverseHashTable.query(verbose=0)
        else:
            candidates = []
            for key in keys:
                candidates += ReverseHashTable.query(key, verbose=0)
        resolved = ub.ddict(list)
        for cand in candidates:
            resolved[cand['key']].extend(
                [f['data'] for f in cand['found']])
        for k in resolved.keys():
            if len(resolved[k]) == 1:
                resolved[k] = resolved[k][0]
        return resolved

    def load1(reporter):
        """
        Load basic data
        """
        table = reporter.state.evaluation_table()
        reporter.state.summarize()
        evaluations = table[~table['raw'].isnull()]
        reporter.raw_df = raw_df = pd.DataFrame(evaluations)

        if 0:
            col_stats_df = unique_col_stats(raw_df)
            print('Column Unique Value Frequencies')
            print(col_stats_df.to_string())

        # test_dset_freq = raw_df['test_dset'].value_counts()
        # print(f'test_dset_freq={test_dset_freq}')

        print('\nRaw')
        num_files_summary(raw_df)

        # Remove duplicate predictions on effectively the same dataset.
        # reporter.filt_df = filt_df = deduplicate_test_datasets(raw_df)
        reporter.raw_df = filt_df = raw_df

        # print('\nDeduplicated (over test dataset)')
        # num_files_summary(filt_df)

        USE_COMP = 0
        if USE_COMP:
            eval_types_to_locs = ub.ddict(list)
            for k, group in filt_df.groupby(['test_dset', 'model', 'pred_cfg']):
                eval_types = tuple(sorted(group['type'].unique()))
                eval_types_to_locs[eval_types].extend(group.index)
            print('Cross-Metric Comparable Locs')
            print(ub.urepr(ub.map_vals(len, eval_types_to_locs)))
            comparable_locs = list(ub.flatten(v for k, v in eval_types_to_locs.items() if len(k) > 0))
            reporter.comp_df = comp_df = filt_df.loc[comparable_locs]

            print('\nCross-Metric Comparable')
            num_files_summary(comp_df)

    def load2(reporter):
        """
        Load detailed data that might cross reference files
        """
        df = reporter.raw_df
        expt_dvc_dpath = reporter.expt_dvc_dpath
        big_rows = reporter.big_rows = load_extended_data(df, expt_dvc_dpath)
        # reporter.big_rows = load_extended_data(reporter.comp_df, reporter.expt_dvc_dpath)
        # set(r['expt'] for r in reporter.big_rows)
        big_rows = reporter.big_rows
        cleaned_df = clean_loaded_data(big_rows, expt_dvc_dpath)
        # reporter.other = other
        reporter.orig_cleaned_df = cleaned_df

        if 0:
            # Remove non-varying (interesting) columns to make development more
            # sane
            keep_cols = []
            for col in cleaned_df.columns:
                keep = True
                try:
                    unique_vals = cleaned_df[col].unique()
                    unique_vals = unique_vals[~pd.isnull(unique_vals)]
                    if len(unique_vals) == 1:
                        if '.' in col:
                            keep = False
                except Exception:
                    pass
                if keep:
                    keep_cols.append(col)
            cleaned_df = cleaned_df[keep_cols]

        if 1:
            # Merge rows from earlier pipeline steps to all of the descendant
            # rows that depend on it.
            # from watch.utils import util_param_grid
            # colnames = ub.oset(cleaned_df.columns)
            # column_nestings = util_dotdict.dotkeys_to_nested(colnames)
            # non_nested = [k for k, v in column_nestings.items() if k == v]
            # column_nestings['trk']
            # column_nestings['act']

            type_to_idkeys = {
                'eval_trk_pxl_fpath': ['trk_model', 'test_trk_dset', 'trk_pxl_cfg'],
                'eval_trk_poly_fpath': ['trk_model', 'test_trk_dset', 'trk_pxl_cfg', 'trk_poly_cfg'],
            }

            metric_names = [c for c in cleaned_df.columns if 'metrics.' in c]
            metric_cols = (ub.oset(metric_names) & cleaned_df.columns)

            def link_types(smaller_row_type, larger_row_type, col_prefix):
                move_cols = [c for c in metric_cols if col_prefix in c]
                if move_cols:
                    smaller_keys = type_to_idkeys[smaller_row_type]
                    smaller = cleaned_df[cleaned_df.type == smaller_row_type]
                    larger = cleaned_df[cleaned_df.type == larger_row_type]
                    new_larger = my_nonstandard_merge(smaller, larger, smaller_keys, move_cols)
                    cleaned_df.loc[new_larger.index, move_cols] = new_larger.loc[:, move_cols].values

            smaller_row_type = 'eval_trk_pxl_fpath'
            larger_row_type = 'eval_trk_poly_fpath'
            col_prefix = 'trk.pxl'
            link_types(smaller_row_type, larger_row_type, col_prefix)

            smaller_row_type = 'eval_trk_pxl_fpath'
            larger_row_type = 'eval_act_poly_fpath'
            col_prefix = 'trk.pxl'
            link_types(smaller_row_type, larger_row_type, col_prefix)
            # move_cols = [c for c in metric_cols if col_prefix in c]
            # smaller_keys = type_to_idkeys[smaller_row_type]
            # smaller = cleaned_df[cleaned_df.type == smaller_row_type]
            # larger = cleaned_df[cleaned_df.type == larger_row_type]
            # new_larger = my_nonstandard_merge(smaller, larger, smaller_keys, move_cols)
            # cleaned_df.loc[new_larger.index, move_cols] = new_larger.loc[:, move_cols].values

            smaller_row_type = 'eval_trk_poly_fpath'
            larger_row_type = 'eval_act_poly_fpath'
            col_prefix = 'trk.poly'
            link_types(smaller_row_type, larger_row_type, col_prefix)
            # move_cols = [c for c in metric_cols if col_prefix in c]
            # smaller_keys = type_to_idkeys[smaller_row_type]
            # smaller = cleaned_df[cleaned_df.type == smaller_row_type]
            # larger = cleaned_df[cleaned_df.type == larger_row_type]
            # new_larger = my_nonstandard_merge(smaller, larger, smaller_keys, move_cols)
            # cleaned_df.loc[new_larger.index, move_cols] = new_larger.loc[:, move_cols].values

            # larger[larger['type'] == 'eval_act_poly_fpath']
            cleaned_df[cleaned_df['type'] == 'eval_trk_pxl_fpath'][metric_cols]
            cleaned_df[cleaned_df['type'] == 'eval_act_poly_fpath'][metric_cols].T

        reporter.orig_merged_df = cleaned_df

        # import xdev
        # xdev.search_replace
        # xdev.set_overlaps(a1['trk_poly_id'], a2['trk_poly_id'])

        # hard coded values
        human_mapping = {
            'coi_mAP': 'Pixelwise mAP (classes of interest)',
            'coi_mAUC': 'Pixelwise mAUC (classes of interest)',
            'salient_AP': 'Pixelwise Salient AP',
            'salient_AUC': 'Pixelwise Salient AUC',
            'sc_macro_f1': 'IARPA SC mean F1',
            'BAS_F1': 'IARPA BAS F1',
            'act_cfg': 'SC Tracking Config',
            'trk_use_viterbi': 'Viterbi Enabled',
            'trk_thresh': 'SC Tracking Threshold',
            'co2_kg': 'CO2 Emissions (kg)',
            'total_hours': 'Time (hours)',
            'sensorchan': 'Sensor/Channel',
            'has_teamfeat': 'Has Team Features',
            'eval_act+pxl': 'SC',
            'eval_trk+pxl': 'BAS',
            'pred_input_space_scale': 'Input Scale',
            'pred_use_cloudmask': 'Cloudmask',
            'pred_resample_invalid_frames': 'Resample Invalid Frames',
            'pred_window_space_scale': 'Window Scale',
        }
        reporter.human_mapping = human_mapping
        reporter.iarpa_metric_lut = {
            'eval_act+pxl': 'sc_macro_f1',
            'eval_trk+pxl': 'BAS_F1',
        }
        reporter.pixel_metric_lut = {
            'eval_act+pxl': 'coi_mAP',
            'eval_trk+pxl': 'salient_AP',
        }
        # reporter.actcfg_to_label = other['actcfg_to_label']
        # reporter.predcfg_to_label = other['predcfg_to_label']
        # reporter.human_mapping.update(reporter.actcfg_to_label)
        # reporter.human_mapping.update(reporter.predcfg_to_label)

    def load(reporter):
        reporter.load1()
        reporter.load2()

    def plot(reporter):
        plot_merged(reporter)


def plot_merged(reporter):
    from watch.mlops.plots import plot_pixel_ap_verus_auc
    from watch.mlops.plots import plot_pixel_ap_verus_iarpa
    from watch.mlops.plots import plot_resource_versus_metric

    reporter = reporter
    dpath = reporter.dpath
    orig_merged_df = reporter.orig_merged_df
    iarpa_metric_lut = reporter.iarpa_metric_lut
    pixel_metric_lut = reporter.pixel_metric_lut
    # predcfg_to_label = reporter.predcfg_to_label
    # actcfg_to_label = reporter.actcfg_to_label
    human_mapping = reporter.human_mapping

    # ['trk_thresh',
    #  'trk_morph_kernel',
    #  'trk_agg_fn',
    #  'trk_thresh_hysteresis',
    #  'trk_moving_window_size']

    common_plotkw = {
        # 'connect': 'expt',
        # 'mesh': 'model',
        # 'clique': 'model',
        'style': 'has_teamfeat',
        # 'star': 'in_production',
        'starkw': {'s': 500},
        's': 120,
    }

    merged_df = orig_merged_df.copy()

    # describe_varied(merged_df, dpath, human_mapping=human_mapping)

    # plot_ta1_vs_l1(merged_df, human_mapping, iarpa_metric_lut, pixel_metric_lut, common_plotkw, dpath)

    plot_pixel_ap_verus_auc(merged_df, human_mapping, iarpa_metric_lut, pixel_metric_lut, common_plotkw, dpath)

    plot_pixel_ap_verus_iarpa(merged_df, human_mapping, iarpa_metric_lut, pixel_metric_lut, common_plotkw, dpath)

    plot_resource_versus_metric(merged_df, human_mapping, iarpa_metric_lut, pixel_metric_lut, common_plotkw, dpath)

    # plot_viterbii_analysis(merged_df, human_mapping, iarpa_metric_lut, pixel_metric_lut, common_plotkw, dpath)


def num_files_summary(df):
    expt_group = dict(list(df.groupby('dataset_code')))
    filt_summaries = []
    groups = sorted(expt_group.items())[::-1]
    for dataset_code, group in groups:
        # print('dataset_code = {!r}'.format(dataset_code))
        # print('Column Unique Value Frequencies')
        # col_stats_df2 = unique_col_stats(group)
        # print(col_stats_df2.to_string())
        row = {}
        row['dataset_code'] = dataset_code
        if 'trk_model' in df.columns:
            row['num_trk_models'] = len(df['trk_model'].dropna().unique())
        else:
            row['num_trk_models'] = 0
        if 'act_model' in df.columns:
            row['num_act_models'] = len(df['act_model'].dropna().unique())
        else:
            row['num_act_models'] = 0
        row.update(
            ub.udict(df['type'].value_counts().to_dict()).map_keys(
                lambda x: 'num_' + x)
        )
        # row['num_experiments'] = len(group['expt'].unique())
        # row['num_models'] = len(group['model'].unique())
        # row['num_pxl_evals'] = type_evals.get('eval_pxl', 0)
        # row['num_bas_evals'] = type_evals.get('eval_trk', 0)
        # row['num_sc_evals'] = type_evals.get('eval_act', 0)
        filt_summaries.append(row)
    _summary_df = pd.DataFrame(filt_summaries)
    total_row = _summary_df.sum().to_dict()
    total_row['dataset_code'] = '*'
    summary_df = pd.DataFrame(filt_summaries + [total_row])
    print('Number of Models & Evaluations')
    print(summary_df.to_string(index=False))


def unique_col_stats(df):
    col_stats = ub.ddict(dict)
    import kwarray
    for key in df.columns:
        col_freq = np.array(list(ub.dict_hist(df[key]).values()))
        stats = kwarray.stats_dict(col_freq, median=True)
        stats['num_unique'] = stats.pop('shape')[0]
        col_stats[key] = stats
        # ['num_unique'] = len(unique_cols)
        # col_stats[key]['max'] = stats['max']
        # col_stats[key]['max'] = stats['max']
    col_stats_df = pd.DataFrame(col_stats)
    # Hack
    col_stats_df = col_stats_df.drop(['dataset_code', 'dvc', 'raw'], axis=1)
    col_stats_df = col_stats_df.astype(int)
    return col_stats_df


def load_extended_data(df, expt_dvc_dpath):
    """
    """
    from watch.mlops import smart_result_parser as frp
    rows = df.to_dict('records')
    big_rows = []
    errors = []

    import os
    WATCH_EVAL_LOAD_STRICT = os.environ.get('WATCH_EVAL_LOAD_STRICT', 1)
    for row in ub.ProgIter(rows, desc='load'):
        big_row = row.copy()
        fpath = row['raw']
        try:
            if row['type'] == 'eval_trk_pxl_fpath':
                pxl_info = frp.load_pxl_eval(fpath, expt_dvc_dpath, arg_prefix='trk.')
                big_row['info'] = pxl_info
            elif row['type'] == 'eval_act_pxl_fpath':
                pxl_info = frp.load_pxl_eval(fpath, expt_dvc_dpath, arg_prefix='act.')
                big_row['info'] = pxl_info
            elif row['type'] == 'eval_act_poly_fpath':
                sc_info = frp.load_sc_poly_eval(fpath, expt_dvc_dpath)
                big_row['info'] = sc_info
            elif row['type'] == 'eval_trk_poly_fpath':
                bas_info = frp.load_bas_poly_eval(fpath, expt_dvc_dpath)
                big_row['info'] = bas_info
            else:
                raise KeyError('Unknown row type: ' + str(row['type']))
            big_rows.append(big_row)
        except Exception as ex:
            errors.append((ex, row))
            if WATCH_EVAL_LOAD_STRICT:
                raise

    import rich
    if len(errors):
        rich.print('[red] ' + repr(errors[0]))
        rich.print(f'[red] {len(errors)=}')
    else:
        print(f'{len(errors)=}')
    return big_rows


def clean_loaded_data(big_rows, expt_dvc_dpath):
    """
    Turn the nested "loaded" data into flat data for tabulation.
    Also combine eval types together into a single row per model / config.
    """

    def fix_none(v):
        return "None" if v is None else v

    # from watch.utils.util_param_grid import dotdict_to_nested
    import kwcoco

    from watch.mlops import expt_manager
    state = expt_manager.ExperimentState(expt_dvc_dpath, '*')

    simple_rows = []
    if 0:
        big_row = big_rows[10]
        big_row = big_rows[-1]

    for big_row in ub.ProgIter(big_rows, desc='big rows'):
        # fpath = big_row['raw']
        row = ub.udict(big_row) - {'info'}
        info = big_row['info']
        param_types = info['param_types']
        params = ub.udict().union(*param_types.values())
        params = params.map_values(fix_none)

        ADD_CROPID_HACK = 1
        if ADD_CROPID_HACK:
            from watch.utils import util_path
            # special handling for adding tracking / cropping
            # params to the activity row. We should figure out a
            # way of making this more general in the future.
            if row['type'] == 'eval_act_poly_fpath':
                if row['test_act_dset'].startswith('crop'):
                    # Fixme dataset name ids need a rework
                    crop_id = row['test_act_dset'].split('_crop.kwcoco')[0]
                    # There needs to be a search step for the crop
                    # dataset, which is not ideal.
                    pats = state.patterns.copy()
                    pats['crop_id'] = crop_id
                    pats = ub.udict(pats).map_values(str)
                    pat = state.templates['crop_fpath'].format(**pats)
                    _found = util_path.coerce_patterned_paths(pat)
                    if _found:
                        assert len(_found) == 1, 'should not have dups here'
                        found = _found[0]
                        _crop_attrs = ub.udict(state._parse_pattern_attrs(state.templates['crop_fpath'], found))
                        _crop_attrs = _crop_attrs - row
                        print(f'_crop_attrs={_crop_attrs}')
                        row.update(_crop_attrs)
            if row['type'] == 'eval_trk_poly_fpath':
                ...

            # Some of the ids from the experiments state may not be build we
            # should do that.
            for k, vs in state.hashid_dependencies.items():
                if k not in row:
                    deps = row & vs
                    if not any(pd.isnull(_) for _ in deps.values()):
                        v = state._condense_cfg(deps, k)
                        row[k] = v

            if row['type'] == 'eval_act_poly_fpath':
                try:
                    if row['regions_id'].startswith('trk_poly_id'):
                        row['trk_poly_id'] = row['regions_id']
                except KeyError:
                    ...

        FIX_FOR_POSTLOAD_TRK_INFO = 1
        if FIX_FOR_POSTLOAD_TRK_INFO:
            extra_attrs = info['other'].get('extra_attrs', None)
            if extra_attrs is None:
                extra_attrs = {}
            extra = ub.udict(extra_attrs) - {k for k, v in row.items() if not pd.isnull(v)}
            row.update(extra)

        for k, v in params.items():
            if k.endswith('.channels'):
                k3 = k.replace('channels', 'has_teamfeat')
                request_sensorchan = kwcoco.SensorChanSpec.coerce(smart.shrink_channels(v))
                row[k3] = smart.is_teamfeat(request_sensorchan)

        row.update(params)
        # row.update(info['metrics'])

        # fit_params = param_type['fit']
        # pred_params = param_type['pred']
        # track_params = info['param_types'].get('track', {})

        # # Shrink and check the sensorchan spec
        # request_sensorchan = kwcoco.SensorChanSpec.coerce(
        #     frp.shrink_channels(fit_params['channels']))
        # fit_params['channels'] = request_sensorchan.spec
        # sensorchan = request_sensorchan

        # fit_params['sensorchan'] = sensorchan
        # row['has_teamfeat'] = _is_teamfeat(sensorchan)

        # fit_param_keys2 = list(fit_param_keys) + ['channels']
        # selected_fit_params = ub.dict_isect(fit_params, fit_param_keys2)
        # row.update(selected_fit_params)
        simple_rows.append(row)

    cleaned_df = pd.DataFrame(simple_rows)
    print(f'{cleaned_df.shape=}')
    # simple_df['sensorchan'].unique()
    # simple_df[simple_df['sensorchan'].isnull()]

    if 0:
        sensorchan_keys = [k for k in cleaned_df.keys() if 'sensorchan' in k]
        print(f'sensorchan_keys={sensorchan_keys}')
        chan_keys = [k for k in cleaned_df.keys() if 'channels' in k]
        print(f'chan_keys={chan_keys}')

    # for gkey, group in cleaned_df.groupby('trk_pxl_cfg'):
    #     pass
    return cleaned_df


def is_null(x):
    return (isinstance(x, float) and math.isnan(x)) or x is None or not bool(x)


def resolve_model_info(model_fpath):
    cacher = ub.Cacher('model_info_memo', depends=[str(model_fpath)], appname='watch/model_info_memo')
    stats = cacher.tryload()
    if stats is None:
        from watch.cli.torch_model_stats import torch_model_stats
        stats = torch_model_stats(model_fpath)
        cacher.save(stats)
    return stats


def checkpoint_filepath_info(fname):
    """
    Finds information encoded in the checkpoint/model file path.

    hacky

    TODO:
        We need to ensure this info is encoded inside the file header as well!

    Ignore
        parse.parse('{prefix}foo={bar}', 'foo=3')
        parse.parse('{prefix}foo={bar}', 'afoao=3')

    Example:
        >>> fnames = [
        >>>     'epoch=1-step=10.foo',
        >>>     'epoch=1-step=10-v2.foo',
        >>>     'epoch=1-step=10',
        >>>     'epoch=1-step=10-v2',
        >>>     'junkepoch=1-step=10.foo',
        >>>     'junk/epoch=1-step=10-v2.foo',
        >>>     'junk-epoch=1-step=10',
        >>>     'junk_epoch=1-step=10-v2',
        >>> ]
        >>> for fname in fnames:
        >>>     info = checkpoint_filepath_info(fname)
        >>>     print(f'info={info}')
        info={'epoch': 1, 'step': 10, 'ckpt_ver': 'v0'}
        info={'epoch': 1, 'step': 10, 'ckpt_ver': 'v2'}
        info={'epoch': 1, 'step': 10, 'ckpt_ver': 'v0'}
        info={'epoch': 1, 'step': 10, 'ckpt_ver': 'v2'}
        info={'epoch': 1, 'step': 10, 'ckpt_ver': 'v0'}
        info={'epoch': 1, 'step': 10, 'ckpt_ver': 'v2'}
        info={'epoch': 1, 'step': 10, 'ckpt_ver': 'v0'}
        info={'epoch': 1, 'step': 10, 'ckpt_ver': 'v2'}
    """
    import parse
    # We assume it must have this
    suffix = ''.join(fname.partition('epoch=')[1:])
    # Hack: making name assumptions
    parsers = [
        parse.Parser('epoch={epoch:d}-step={step:d}-{ckpt_ver}.{ext}'),
        parse.Parser('epoch={epoch:d}-step={step:d}.{ext}'),
        parse.Parser('epoch={epoch:d}-step={step:d}-{ckpt_ver}'),
        parse.Parser('epoch={epoch:d}-step={step:d}'),
    ]
    # results = parser.parse(str(path))
    info = None
    for parsers in parsers:
        result = parsers.parse(suffix)
        if result:
            break
    if result:
        info = result.named
        if 'ckpt_ver' not in info:
            info['ckpt_ver'] = 'v0'
        info = ub.dict_diff(info, {'ext', 'prefix'})
    return info


def my_nonstandard_merge(smaller, larger, smaller_keys, move_cols, mode=0):
    """
    We are copying specific columns from a single row in a smaller dataframe
    into multiple rows in a larger data frame.

    Args:
        smaller (pd.DataFrame): a data frame to copy from
        larger (pd.DataFrame): a data frame to copy into
        smaller_keys (List[str]): columns that specify a single row in
            `smaller` and groups of rows in `larger`
        move_cols (List[str]): the columns to move.

    Example:
        >>> from watch.mlops.old.expt_report import *  # NOQA
        >>> smaller = pd.DataFrame([
        >>>     {'k1': 1, 'k2': 1, 'm1': 2, 'm2': 2, 's': 2, 'u1': 9},
        >>>     {'k1': 3, 'k2': 3, 'm1': 4, 'm2': 3, 's': 2, 'u2': 8},
        >>>     {'k1': 5, 'k2': 5, 'm1': 6, 'm2': 5, 's': 2, 'u3': 7},
        >>> ])
        >>> larger = pd.DataFrame([
        >>>     {'k1': 1, 'k2': 1, 'm1': np.nan, 'u2': 1, 's': 3},
        >>>     {'k1': 1, 'k2': 1, 'm1': np.nan, 'u2': 2, 's': 3},
        >>>     {'k1': 3, 'k2': 3, 'm1': np.nan, 'u2': 3, 's': 3},
        >>>     {'k1': 3, 'k2': 3, 'm1': np.nan, 'u2': 4, 's': 3},
        >>>     {'k1': 5, 'k2': 5, 'm1': np.nan, 'u2': 5, 's': 3},
        >>>     {'k1': 5, 'k2': 5, 'm1': np.nan, 'u2': 6, 's': 3},
        >>> ])
        >>> smaller_keys = ['k1', 'k2']  # should be usable as an index
        >>> move_cols = ['m1', 'm2']  # columns to move
        >>> larger1 = my_nonstandard_merge(smaller, larger, smaller_keys, move_cols, mode=0)
        >>> print(larger1)
        >>> larger2 = my_nonstandard_merge(smaller, larger, smaller_keys, move_cols, mode=1)
        >>> print(larger2)

    Ignore:
        import timerit
        ti = timerit.Timerit(100, bestof=10, verbose=2)
        for timer in ti.reset('time'):
            my_nonstandard_merge(smaller, larger, smaller_keys, move_cols, mode=1)
        for timer in ti.reset('time'):
            my_nonstandard_merge(smaller, larger, smaller_keys, move_cols, mode=0)
    """
    if mode == 0:
        smaller_suffix = None
        larger_suffix = '_y'
        smaller_subset = smaller[smaller_keys + move_cols]
        new_larger = pd.merge(smaller_subset, larger, on=smaller_keys,
                              suffixes=[smaller_suffix, larger_suffix],
                              validate='one_to_many')
        drop_cols = [k + larger_suffix for k in larger.columns.intersection(move_cols)]
        new_larger = new_larger.drop(drop_cols, axis=1)
        return new_larger
    else:
        smaller_lut = smaller.set_index(smaller_keys)
        new_larger = larger.copy()
        for smaller_key, group in dict(list(larger.groupby(smaller_keys))).items():
            small_match = smaller_lut.loc[smaller_key, move_cols]
            # Why is .values needed here? TODO: understand
            new_larger.loc[group.index, move_cols] = small_match.values
            # larger.loc[group.index, move_cols] = small_match
        return new_larger


# def deduplicate_test_datasets(raw_df):
#     """
#     The same model might have been run on two variants of the dataset.
#     E.g. a RGB model might have run on data_vali.kwcoco.json and
#     combo_DILM.kwcoco.json. The system sees these as different datasets
#     even though the model will use the same subset of both. We define
#     a heuristic ordering and then take just one of them.
#     """
#     preference = {
#         'Cropped-Drop3-TA1-2022-03-10_combo_DLM_s2_wv_vali.kwcoco': 0,
#         'Cropped-Drop3-TA1-2022-03-10_combo_DL_s2_wv_vali.kwcoco': 1,
#         'Cropped-Drop3-TA1-2022-03-10_data_wv_vali.kwcoco': 2,
#         'Aligned-Drop3-TA1-2022-03-10_combo_LM_nowv_vali.kwcoco': 0,
#         'Aligned-Drop3-TA1-2022-03-10_combo_LM_vali.kwcoco': 1,
#     }
#     FILTER_DUPS = 1
#     if FILTER_DUPS:
#         keep_locs = []
#         for k, group in raw_df.groupby(['dataset_code', 'model', 'pred_cfg', 'type']):
#             prefs = group['test_dset'].apply(lambda x: preference.get(x, 0))
#             keep_flags = prefs == prefs.min()
#             keep_locs.extend(group[keep_flags].index)
#         print(f'Keep {len(keep_locs)} / {len(raw_df)} drop3 evals')
#         filt_df = raw_df.loc[keep_locs]
#     else:
#         filt_df = raw_df.copy()
#     return filt_df


def get_column_meanings():
    return [
        {'name': 'raw', 'help': 'A full path to a file on disk that contains this info'},
        {'name': 'dvc', 'help': 'A path to a DVC sidecar file if it exists.'},
        {'name': 'type', 'help': 'The type of the row'},
        {'name': 'step', 'help': 'The number of steps taken by the most recent training run associated with the row'},
        {'name': 'total_steps', 'help': 'An estimate of the total number of steps the model associated with the row took over all training runs.'},
        {'name': 'model', 'help': 'The name of the learned model associated with this row'},
        # {'name': 'test_dset', 'help': 'The name of the test dataset used to compute a metric associated with this row'},
        {'name': 'test_trk_dset', 'help': 'The name of the test BAS dataset used to compute a metric associated with this row'},
        {'name': 'test_act_dset', 'help': 'The name of the test SC dataset used to compute a metric associated with this row'},

        {'name': 'expt', 'help': 'The name of the experiment, i.e. training session that might have made several models'},
        {'name': 'dataset_code', 'help': 'The higher level dataset code associated with this row'},

        {'name': 'pred_cfg', 'help': 'A hash of the configuration used for pixel heatmap prediction'},
        {'name': 'trk_cfg', 'help': 'A hash of the configuration used for BAS tracking'},
        {'name': 'act_cfg', 'help': 'A hash of the configuration used for SC classification'},

        {'name': 'total_steps', 'help': 'An estimate of the total number of steps the model associated with the row took over all training runs.'},
    ]



import os
import ubelt as ub
import pathlib
from watch.utils import util_pattern


def coercepath(path_like):
    """
    Args:
        path_like (str | ub.Path | os.PathLike):
            an object representing a filesystem path

    DEPRECATE:
        you canjust call ub.Path instead

    Example:
        >>> from watch.utils.util_path import *  # NOQA
        >>> #
        >>> path_like = '.'
        >>> path = coercepath(path_like)
        >>> print('path = {!r}'.format(path))
        >>> #
        >>> path_like = ub.Path('.')
        >>> path = coercepath(path_like)
        >>> print('path = {!r}'.format(path))
        >>> #
        >>> path_like = pathlib.PurePath('.')
        >>> path = coercepath(path_like)
        >>> print('path = {!r}'.format(path))
    """
    if isinstance(path_like, str):
        path = ub.Path(path_like)
    elif isinstance(path_like, pathlib.PurePath):
        path = path_like
    elif isinstance(path_like, os.PathLike):
        path = path_like
    else:
        raise TypeError('Unable to coerce {} to Path'.format(type(path_like)))
    return path


def tree(path):
    """
    Like os.walk but yields a flat list of file and directory paths

    Args:
        path (str | os.PathLike)

    Yields:
        str: path

    Example:
        >>> import itertools as it
        >>> from watch.utils.util_path import *  # NOQA
        >>> import ubelt as ub
        >>> path = ub.Path('.')
        >>> gen = tree(path)
        >>> results = list(it.islice(gen, 5))
        >>> print('results = {}'.format(ub.urepr(results, nl=1)))
    """
    import os
    from os.path import join
    for r, fs, ds in os.walk(path):
        for f in fs:
            yield join(r, f)
        for d in ds:
            yield join(r, d)


def coerce_patterned_paths(data, expected_extension=None):
    """
    Coerce input to a list of paths.

    Args:
        data (str | List[str]):
            a glob pattern or list of glob patterns or a yaml list of glob
            patterns

    Returns:
        List[ubelt.Path]: Multiple paths that match the query

    Example:
        >>> # xdoctest: +SKIP
        >>> import watch
        >>> dvc_dpath = watch.find_dvc_dpath(tags='phase2_data')
        >>> data = [dvc_dpath / 'annotations/region_models/*.geojson']
        >>> from watch.utils.util_path import *  # NOQA
        >>> print(coerce_patterned_paths(dvc_dpath / 'annotations/region_models'))
        >>> print(coerce_patterned_paths([dvc_dpath / 'annotations/region_models/*.geojson']))
        >>> print(coerce_patterned_paths('./'))

        import watch
        dvc_dpath = watch.find_dvc_dpath(tags='phase2_data')
        data = [ub.Path(dvc_dpath / 'annotations/region_models')]
        expected_extension = ['*.geojson', '*.txt']

    Example:
        >>> import watch
        >>> empty_fpaths = coerce_patterned_paths(None)
        >>> assert len(empty_fpaths) == 0

    Example:
        >>> from watch.utils.util_path import *  # NOQA
        >>> import watch
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('watch/test/utils/path/').ensuredir()
        >>> (dpath / 'file1.txt').touch()
        >>> (dpath / 'dir').ensuredir()
        >>> (dpath / 'dir' / 'subfile1.txt').touch()
        >>> (dpath / 'dir' / 'subfile2.txt').touch()
        >>> paths = coerce_patterned_paths(
        ...     f'''
        ...     - {dpath / 'file1.txt'}
        ...     - {dpath / 'file2.txt'}
        ...     - {dpath / 'dir'}
        ...     ''', expected_extension='.txt')
        >>> paths = [p.shrinkuser() for p in paths]
        >>> print('paths = {}'.format(ub.urepr(paths, nl=1)))
        >>> with ChDir(dpath / 'dir'):
        >>>     paths = coerce_patterned_paths('*.txt*')
        >>> print('paths = {}'.format(ub.urepr(paths, nl=1)))
        >>> assert len(paths) == 2

        paths = [
            Path('~/.cache/watch/test/utils/path/file1.txt'),
            Path('~/.cache/watch/test/utils/path/dir/subfile1.txt'),
            Path('~/.cache/watch/test/utils/path/dir/subfile2.txt'),
        ]
    """
    from watch.utils.util_yaml import Yaml
    from ruamel.yaml.composer import ComposerError
    from os.path import isdir, join
    import glob

    if data is None:
        datas = []
    elif ub.iterable(data):
        datas = data
    else:
        datas = [data]

    # Resolve any yaml
    resolved_globs = []
    for data in datas:
        if isinstance(data, str):
            try:
                loaded = Yaml.loads(data)
            except ComposerError:
                loaded = data
            if isinstance(loaded, str):
                loaded = [loaded]
            resolved_globs.extend(loaded)
        else:
            resolved_globs.append(data)

    datas = list(map(os.fspath, resolved_globs))
    paths = []
    for data_ in resolved_globs:
        if expected_extension is not None and isdir(data_):
            exts = expected_extension if ub.iterable(expected_extension) else [expected_extension]
            globpats = [join(data_, '*' + e) for e in exts]
        else:
            globpats = [data_]
        for globpat in globpats:
            paths.extend(list(glob.glob(os.fspath(globpat), recursive=True)))
    paths = [ub.Path(p) for p in paths]
    return paths


def find(pattern=None, dpath=None, include=None, exclude=None, type=None,
         recursive=True, followlinks=False):
    """
    Find all paths in a root subject to a search criterion

    Args:
        pattern (str):
            The glob pattern the path name must match to be returned

        dpath (str):
            The root direcotry to search. Default to cwd.

        include (str | List[str]):
            Pattern or list of patterns. If specified, search only files whose
            base name matches this pattern. By default the pattern is GLOB.

        exclude (str | List[str]):
            Pattern or list of patterns. Skip any file with a name suffix that
            matches the pattern. By default the pattern is GLOB.

        type (str | List[str]):
            A list of 1 character codes indicating what types of file can be
            returned. Currently we only allow either "f" for file or "d" for
            directory. Symbolic links are not currently distinguished. In the
            future we may support posix codes, see [1]_ for details.

        recursive:
            search all subdirectories recursively

        followlinks (bool, default=False):
            if True will follow directory symlinks

    References:
        _[1] https://linuxconfig.org/identifying-file-types-in-linux

    TODO:
        mindepth

        maxdepth

        ignore_case

        regex_match


    Example:
        >>> from watch.utils.util_path import *  # NOQA
        >>> paths = list(find(pattern='*'))
        >>> paths = list(find(pattern='*', type='f'))
        >>> print('paths = {!r}'.format(paths))
        >>> print('paths = {!r}'.format(paths))
    """
    from os.path import join

    if pattern is None:
        pattern = '*'

    if type is None:
        with_dirs = True
        with_files = True
    else:
        with_dirs = False
        with_files = False
        if 'd' in type:
            with_dirs = True
        if 'f' in type:
            with_files = True

    if dpath is None:
        dpath = os.getcwd()

    include_ = (None if include is None else
                util_pattern.MultiPattern(include, hint='glob'))
    exclude_ = (None if exclude is None else
                util_pattern.MultiPattern(exclude, hint='glob'))

    main_pattern = util_pattern.Pattern.coerce(pattern, hint='glob')

    def is_included(name):
        if not main_pattern.match(name):
            return False

        if exclude_ is not None:
            if exclude_.match(name):
                return False

        if include_ is not None:
            if include_.match(name):
                return True
            else:
                return False
        return True

    for root, dnames, fnames in os.walk(dpath, followlinks=followlinks):

        if with_files:
            for fname in fnames:
                if is_included(fname):
                    yield join(root, fname)

        if with_dirs:
            for dname in dnames:
                if is_included(dname):
                    yield join(root, dname)

        if not recursive:
            break


def resolve_relative_to(path, dpath, strict=False):
    """
    Given a path, try to resolve its symlinks such that it is relative to the
    given dpath.

    Ignore:
        def _symlink(self, target, verbose=0):
            return ub.Path(ub.symlink(target, self, verbose=verbose))
        ub.Path._symlink = _symlink

        # TODO: try to enumerate all basic cases

        base = ub.Path.appdir('kwcoco/tests/reroot')
        base.delete().ensuredir()

        drive1 = (base / 'drive1').ensuredir()
        drive2 = (base / 'drive2').ensuredir()

        data_repo1 = (drive1 / 'data_repo1').ensuredir()
        cache = (data_repo1 / '.cache').ensuredir()
        real_file1 = (cache / 'real_file1').touch()

        real_bundle = (data_repo1 / 'real_bundle').ensuredir()
        real_assets = (real_bundle / 'assets').ensuredir()

        # Symlink file outside of the bundle
        link_file1 = (real_assets / 'link_file1')._symlink(real_file1)
        real_file2 = (real_assets / 'real_file2').touch()
        link_file2 = (real_assets / 'link_file2')._symlink(real_file2)


        # A symlink to the data repo
        data_repo2 = (drive1 / 'data_repo2')._symlink(data_repo1)
        data_repo3 = (drive2 / 'data_repo3')._symlink(data_repo1)
        data_repo4 = (drive2 / 'data_repo4')._symlink(data_repo2)

        # A prediction repo TODO
        pred_repo5 = (drive2 / 'pred_repo5').ensuredir()

        _ = ub.cmd(f'tree -a {base}', verbose=3)

        fpaths = []
        for r, ds, fs in os.walk(base, followlinks=True):
            for f in fs:
                if 'file' in f:
                    fpath = ub.Path(r) / f
                    fpaths.append(fpath)


        dpath = real_bundle.resolve()

        for path in fpaths:
            # print(f'{path}')
            # print(f'{path.resolve()=}')
            resolved_rel = resolve_relative_to(path, dpath)
            print('resolved_rel = {!r}'.format(resolved_rel))
    """
    try:
        resolved_abs = resolve_directory_symlinks(path)
        resolved_rel = resolved_abs.relative_to(dpath)
    except ValueError:
        if strict:
            raise
        else:
            return path
    return resolved_rel


def resolve_directory_symlinks(path):
    """
    Only resolve symlinks of directories
    """
    return path.parent.resolve() / path.name
    # prev = path
    # curr = prev.parent
    # while prev != curr:
    #     if curr.is_symlink():
    #         rhs = path.relative_to(curr)
    #         resolved_lhs = curr.resolve()
    #         new_path = resolved_lhs / rhs
    #         return new_path
    #     prev = curr
    #     curr = prev.parent
    # return path


class ChDir:
    """
    Context manager that changes the current working directory and then
    returns you to where you were.

    Args:
        dpath (PathLike | None):
            The new directory to work in.
            If None, then the context manager is disabled.

    Example:
        >>> from watch.utils.util_path import *  # NOQA
        >>> dpath = ub.Path.appdir('xdev/tests/chdir').ensuredir()
        >>> dir1 = (dpath / 'dir1').ensuredir()
        >>> dir2 = (dpath / 'dir2').ensuredir()
        >>> with ChDir(dpath):
        >>>     assert ub.Path.cwd() == dpath
        >>>     # changes to the given directory, and then returns back
        >>>     with ChDir(dir1):
        >>>         assert ub.Path.cwd() == dir1
        >>>         with ChDir(dir2):
        >>>             assert ub.Path.cwd() == dir2
        >>>             # changes inside the context manager will be reset
        >>>             os.chdir(dpath)
        >>>         assert ub.Path.cwd() == dir1
        >>>     assert ub.Path.cwd() == dpath
        >>>     with ChDir(dir1):
        >>>         assert ub.Path.cwd() == dir1
        >>>         with ChDir(None):
        >>>             assert ub.Path.cwd() == dir1
        >>>             # When disabled, the cwd does *not* reset at context exit
        >>>             os.chdir(dir2)
        >>>         assert ub.Path.cwd() == dir2
        >>>         os.chdir(dir1)
        >>>         # Dont change dirs, but reset to your cwd at context end
        >>>         with ChDir('.'):
        >>>             os.chdir(dir2)
        >>>         assert ub.Path.cwd() == dir1
        >>>     assert ub.Path.cwd() == dpath
    """

    def __init__(self, dpath):
        self.context_dpath = dpath
        self.orig_dpath = None

    def __enter__(self):
        if self.context_dpath is not None:
            self.orig_dpath = os.getcwd()
            os.chdir(self.context_dpath)
        return self

    def __exit__(self, a, b, c):
        if self.context_dpath is not None:
            os.chdir(self.orig_dpath)


def sidecar_glob(main_pat, sidecar_ext, main_key='main', sidecar_key=None,
                 recursive=0):
    """
    Similar to a regular glob, but returns a dictionary with associated
    main-file / sidecar-file pairs.

    TODO:
        add as a general option to Pattern.paths?

    Args:
        main_pat (str | PathLike):
            glob pattern for the main non-sidecar file

    Yields:
        Dict[str, Path | None]

    Notes:
        A sidecar file is defined by the sidecar extension. We usually use this
        for .dvc sidecars.

        When the pattern includes a .dvc suffix, the result will include those .dvc
        files and any matching main files they correspond to. Note: if you search
        for paths like `foo_*.dvc` this might skiped unstaged files. Therefore it
        is recommended to only include the .dvc suffix in the pattern ONLY if you
        do not want any unstaged files.

        If you want both staged and unstaged files, ensure the pattern does not
        exclude objects without a .dvc suffix (i.e. don't end the pattern with
        .dvc).

        When the pattern does not include a .dvc suffix, we include all those
        files, for other files that exist by adding a .dvc suffix.

        With the pattern matches both a dvc and non-dvc file, they are grouped
        together.

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> import watch
        >>> dvc_dpath = watch.find_smart_dvc_dpath()
        >>> bundle_dpath = dvc_dpath / 'Cropped-Drop3-TA1-2022-03-10'
        >>> sidecar_ext = '.dvc'
        >>> print(ub.urepr(list(sidecar_glob(bundle_dpath / '*R*', sidecar_ext)), nl=2))
        >>> print(ub.urepr(list(sidecar_glob(bundle_dpath / '*.dvc', sidecar_ext)), nl=2))

    Example:
        >>> from watch.utils.util_path import *  # NOQA
        >>> dpath = ub.Path.appdir('xdev/tests/sidecar_glob')
        >>> dpath.delete().ensuredir()
        >>> (dpath / 'file1').touch()
        >>> (dpath / 'file1.ext').touch()
        >>> (dpath / 'file1.ext.car').touch()
        >>> (dpath / 'file2.ext').touch()
        >>> (dpath / 'file3.ext.car').touch()
        >>> (dpath / 'file4.car').touch()
        >>> (dpath / 'file5').touch()
        >>> (dpath / 'file6').touch()
        >>> (dpath / 'file6.car').touch()
        >>> (dpath / 'file7.bike').touch()
        >>> def _handle_resulst(results):
        ...     results = list(results)
        ...     for row in results:
        ...         for k, v in row.items():
        ...             if v is not None:
        ...                 row[k] = v.relative_to(dpath)
        ...     print(ub.urepr(results, sv=1))
        ...     return results
        >>> main_key = 'main',
        >>> sidecar_key = '.car'
        >>> sidecar_ext = '.car'
        >>> main_pat = dpath / '*'
        >>> _handle_resulst(sidecar_glob(main_pat, sidecar_ext))
        >>> _handle_resulst(sidecar_glob(dpath / '*.ext', '.car'))
        >>> _handle_resulst(sidecar_glob(dpath / '*.car', '.car'))
        >>> _handle_resulst(sidecar_glob(dpath / 'file*.ext', '.car'))
        >>> _handle_resulst(sidecar_glob(dpath / '*', '.ext'))
    """
    from watch.utils import util_pattern
    import warnings
    import os
    _len_ext = len(sidecar_ext)
    main_pat = os.fspath(main_pat)
    glob_patterns = [main_pat]
    if main_pat.endswith(sidecar_ext):
        warnings.warn(
            'The main path query should not end with the sidecar extension.'
            ' {main_pat=} {sidecar_ext=}'
        )
        # We could have a variant that removes the extension, but lets not do
        # that and document it.
        # glob_patterns.append(pat[:-_len_ext])
    else:
        if main_pat.endswith('/*'):
            # Optimization dont need an extra pattern in this case
            pass
        else:
            glob_patterns.append(main_pat + sidecar_ext)

    mpat = util_pattern.MultiPattern.coerce(glob_patterns)
    if sidecar_key is None:
        sidecar_key = sidecar_ext
    default = {main_key: None, sidecar_key: None}
    id_to_row = ub.ddict(default.copy)
    paths = mpat.paths(recursive=recursive)

    def _gen():
        for path in paths:
            parent = path.parent
            name = path.name
            if name.endswith(sidecar_ext):
                this_key = sidecar_key
                other_key = main_key
                main_path = parent / name[:-_len_ext]
                other_path = main_path
            else:
                this_key = main_key
                other_key = sidecar_key
                main_path = path
                other_path = parent / (name + sidecar_ext)
            needs_yeild = main_path not in id_to_row
            row = id_to_row[main_path]
            row[this_key] = path
            if row[other_key] is None:
                if other_path.exists():
                    row[other_key] = other_path
            if needs_yeild:
                yield row
    # without this, yilded rows might modify themselves later, that is
    # confusing for a user. Don't do it or come up with a scheme where we
    # detect if a row is "complete" and only yeild it then
    # We could more easilly do this if we used a walk-style find and pattern
    # match mechanism
    rows = list(_gen())
    yield from rows




def parse_delimited_argstr(data):
    """
    Special suffixes can be added to generic demo names. Parse them out here.
    Arguments are `-` separated, only known defaulted values are parsed. Bare
    default names are interpreted as a value of True, otherwise the value
    should be numeric. TODO: generalize this and conslidate in the kwcoco
    demo method.

    Example:
        >>> from watch.utils.util_codes import *  # NOQA
        >>> data = 'foo-bar-baz1-biz2.3'
        >>> defaults = {}
        >>> alias_to_key = None
        >>> parse_delimited_argstr(data)
        {'foo': True, 'bar': True, 'baz': 1, 'biz': 2.3}
    """
    import re
    from scriptconfig.smartcast import smartcast
    parts = data.split('-')
    parsed = {}
    for part in parts:
        match = re.search(r'[\d]', part)
        if match is None:
            value = True
            key = part
        else:
            key = part[:match.span()[0]]
            value = smartcast(part[match.span()[0]:], allow_split=False)
        parsed[key] = value
    return parsed



# Autogenerated via:
# python ~/code/watch/dev/maintain/mirror_package_geowatch.py
def __getattr__(key):
    import watch.tasks.rutgers_material_seg.experiments.iarpa_patch_finetune as mirror
    return getattr(mirror, key)


#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
import scriptconfig as scfg
import warnings
import ubelt as ub


class CropSitesToRegionsConfig(scfg.DataConfig):
    r"""
    Crops site models to the bounds of a region model.

    TODO:
        - [ ] Rename this to TrimSiteModels?

    Example:
        DVC_DPATH=$(WATCH_PREIMPORT=none python -m watch.cli.find_dvc)
        WATCH_PREIMPORT=none python -m watch.cli.crop_sites_to_regions \
            --site_models "$DVC_DPATH/annotations/site_models/KR_R002_*.geojson" \
            --region_models "$DVC_DPATH/annotations/region_models/KR_R002.geojson" \
            --new_site_dpath ./cropped_sites
    """
    __default__ = {
        'site_models': scfg.Value(None, help=ub.paragraph(
            '''
            Geospatial geojson "site" annotation files. Either a path to a
            file, or a directory.
            ''')),

        'region_models': scfg.Value(None, help=ub.paragraph(
            '''
            A single geojson "region" file to crop to.
            ''')),

        'new_site_dpath': scfg.Value(None, help=ub.paragraph(
            '''
            Destination directory for new site models.
            Note: names of files must be unique.
            ''')),

        'new_region_dpath': scfg.Value(None, help=ub.paragraph(
            '''
            Destination directory for new site models.
            Note: names of files must be unique.
            ''')),

        'io_workers': scfg.Value(0, help=ub.paragraph(
            '''
            IO workers to load sites in the background while others are
            cropping.
            ''')),
        'force_multipolygon': scfg.Value(True, help=ub.paragraph(
            '''
            For output site observations the output geometry type will
            be set to MultiPolygon.  As per the T&E specification
            ''')),
    }


USE_LISTS = 0  # turn on for eager debugging


def main(cmdline=False, **kwargs):
    from watch.utils import util_gis
    from shapely.geometry import MultiPolygon
    import geopandas as gpd
    import safer

    config = CropSitesToRegionsConfig.cli(data=kwargs, cmdline=cmdline,
                                          strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1)))

    new_site_dpath = config['new_site_dpath']
    assert new_site_dpath is not None, 'must specify new_site_dpath'
    new_site_dpath = ub.Path(new_site_dpath)

    new_region_dpath = config['new_region_dpath']
    assert new_region_dpath is not None, 'must specify new_region_dpath'
    new_region_dpath = ub.Path(new_region_dpath)

    site_geojson_fpaths: list[ub.Path] = util_gis.coerce_geojson_paths(
        config.get('site_models'))

    region_geojson_fpaths: list[ub.Path] = util_gis.coerce_geojson_paths(
        config.get('region_models'))

    # Load a single region
    if len(region_geojson_fpaths) != 1:
        raise ValueError(f'Must specify exactly one region file, Got: {region_geojson_fpaths}')

    regions = list(util_gis.load_geojson_datas(
        region_geojson_fpaths, workers=0, desc='load geojson region-models'))
    old_region_fpath = regions[0]['fpath']
    region_gdf_crs84: gpd.GeoDataFrame = regions[0]['data']

    # Load multiple site models
    io_workers = config['io_workers']
    sites = util_gis.load_geojson_datas(
        site_geojson_fpaths, workers=io_workers,
        desc='load geojson site-models')

    if USE_LISTS:
        sites = list(sites)

    cropped_region, cropped_sites = crop_sites_to_region(
        region_gdf_crs84, sites)

    if USE_LISTS:
        cropped_sites = ub.ProgIter(
            cropped_sites, desc='Cropping sites',
            total=len(site_geojson_fpaths))
        cropped_sites = list(cropped_sites)

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'pandas.Int64Index is deprecated', FutureWarning)

        # Save the cropped site summaries to disk
        new_region_dpath.ensuredir()
        new_region_fpath = new_region_dpath / old_region_fpath.name

        # Not sure why this insists on bytes. I dont think it was before
        with safer.open(new_region_fpath, temp_file=not ub.WIN32, mode='w') as file:
            cropped_region_json = cropped_region.to_json(na='drop', indent=2)
            file.write(cropped_region_json)
        print(f'Wrote cropped site summaries to {new_region_fpath}')

        # Save the cropped sites to disk
        new_site_dpath.ensuredir()
        num_valid = 0
        total = 0
        for cropped_site_info in cropped_sites:
            total += 1
            cropped_site = cropped_site_info['data']
            if len(cropped_site):
                num_valid += 1
                old_site_fpath = cropped_site_info['fpath']
                new_site_fpath = new_site_dpath / old_site_fpath.name
                cropped_site['observation_date'] = cropped_site['observation_date'].astype('string')
                cropped_site['start_date'] = cropped_site['start_date'].astype('string')
                cropped_site['end_date'] = cropped_site['end_date'].astype('string')

                if config['force_multipolygon']:
                    cropped_site.loc[cropped_site['type'] == 'observation', 'geometry'] =\
                        cropped_site[cropped_site['type'] == 'observation']['geometry'].apply(
                            lambda x: MultiPolygon((x,)) if not isinstance(x, MultiPolygon) else x)

                if 'predicted_phase_transition_date' in cropped_site:
                    cropped_site['predicted_phase_transition_date'] =\
                        cropped_site['predicted_phase_transition_date'].astype('string')

                with safer.open(new_site_fpath, temp_file=not ub.WIN32, mode='w') as file:
                    cropped_site_json = cropped_site.to_json(
                        na='drop', indent=2)
                    file.write(cropped_site_json)
                    # cropped_site.to_file(file, driver='GeoJSON')
        print(f'Wrote {num_valid} / {total} valid cropped sites in {new_site_dpath}')


def crop_sites_to_region(region_gdf_crs84, sites):
    """
    Args:
        region_gdf_crs84 (GeoDataFrame):
            the region GDF containing the region geom to crop to and
            the site summary geometry

        sites (Iterable[Dict]):
            List of the loaded geo data frames with a 'data' key
            and the file path in the 'fpath' key.

    Returns:
        Tuple[GeoDataFrame, List[Dict]]:
            Region model with cropped site summaries and a list of site info
            dictionaries containing the new cropped data field.

    Example:
        >>> from watch.cli.crop_sites_to_regions import *  # NOQA
        >>> import geopandas as gpd
        >>> import kwimage
        >>> from watch.utils import util_gis
        >>> crs84 = util_gis._get_crs84()
        >>> region_poly = kwimage.Polygon.random(rng=0).translate((42, 72))
        >>> site_poly1 = region_poly.translate((0.0001, 0.0001))
        >>> #
        >>> def demo_site_summary(site_id, site_poly):
        >>>     return {
        >>>         'type': 'site_summary',
        >>>         'region_id': None,
        >>>         'site_id': site_id,
        >>>         'start_date': '2020-01-01',
        >>>         'end_date': '2020-01-03',
        >>>         'geometry': site_poly.to_shapely()
        >>>     }
        >>> def demo_site(site_id, site_poly):
        >>>     sh_poly = site_poly.to_shapely()
        >>>     site = gpd.GeoDataFrame([
        >>>         {'type': 'site', 'region_id': 'DemoRegion', 'site_id': site_id, 'geometry': sh_poly},
        >>>         {'type': 'observation', 'observation_date': '2020-01-01', 'current_phase': 'phase1', 'geometry': sh_poly},
        >>>         {'type': 'observation', 'observation_date': '2020-01-02', 'current_phase': 'phase2', 'geometry': sh_poly},
        >>>         {'type': 'observation', 'observation_date': '2020-01-03', 'current_phase': 'phase3', 'geometry': sh_poly},
        >>>     ], crs=crs84)
        >>>     return {'fpath': None, 'data': site}
        >>> region_gdf_crs84 = gpd.GeoDataFrame([
        >>>     {
        >>>         'type': 'region',
        >>>         'region_id': 'DemoRegion',
        >>>         'geometry': region_poly.to_shapely(),
        >>>     },
        >>>     demo_site_summary('DemoRegion_0001', site_poly1),
        >>> ], crs=crs84)
        >>> sites = [
        >>>     demo_site('DemoRegion_0001', site_poly1),
        >>> ]
        >>> cropped_region, cropped_sites = crop_sites_to_region(region_gdf_crs84, sites)
        >>> cropped_sites = list(cropped_sites)
        >>> assert len(cropped_sites) == len(sites)
        >>> assert len(cropped_region) == 2

    Example:
        >>> # xdoctest: +REQUIRES(--slow)
        >>> from watch.cli.crop_sites_to_regions import *  # NOQA
        >>> import geopandas as gpd
        >>> import kwimage
        >>> from watch.utils import util_gis
        >>> crs84 = util_gis._get_crs84()
        >>> region_poly = kwimage.Polygon.random(rng=0).translate((42, 72))
        >>> site_poly0 = region_poly
        >>> site_poly1 = region_poly.translate((0.0001, 0.0001))
        >>> site_poly2 = region_poly.translate((3.0, 3.0))
        >>> site_poly3 = kwimage.Polygon.random().translate((42, 72))
        >>> site_poly4 = kwimage.Polygon.random().translate((43, 73))
        >>> site_poly5 = kwimage.Polygon.random().translate((42.1, 72.1))
        >>> #
        >>> def demo_site_summary(site_id, site_poly):
        >>>     return {
        >>>         'type': 'site_summary',
        >>>         'region_id': None,
        >>>         'site_id': site_id,
        >>>         'start_date': '2020-01-01',
        >>>         'end_date': '2020-01-03',
        >>>         'geometry': site_poly.to_shapely()
        >>>     }
        >>> def demo_site(site_id, site_poly):
        >>>     sh_poly = site_poly.to_shapely()
        >>>     site = gpd.GeoDataFrame([
        >>>         {'type': 'site', 'region_id': 'DemoRegion', 'site_id': site_id, 'geometry': sh_poly},
        >>>         {'type': 'observation', 'observation_date': '2020-01-01', 'current_phase': 'phase1', 'geometry': sh_poly},
        >>>         {'type': 'observation', 'observation_date': '2020-01-02', 'current_phase': 'phase2', 'geometry': sh_poly},
        >>>         {'type': 'observation', 'observation_date': '2020-01-03', 'current_phase': 'phase3', 'geometry': sh_poly},
        >>>     ], crs=crs84)
        >>>     return {'fpath': None, 'data': site}
        >>> region_gdf_crs84 = gpd.GeoDataFrame([
        >>>     {
        >>>         'type': 'region',
        >>>         'region_id': 'DemoRegion',
        >>>         'geometry': region_poly.to_shapely(),
        >>>     },
        >>>     demo_site_summary('DemoRegion_0001', site_poly1),
        >>>     demo_site_summary('DemoRegion_0002', site_poly2),
        >>> ], crs=crs84)
        >>> sites = [
        >>>     demo_site('DemoRegion_0000', site_poly0),
        >>>     demo_site('DemoRegion_0001', site_poly1),
        >>>     demo_site('DemoRegion_0002', site_poly2),
        >>>     demo_site('DemoRegion_0003', site_poly3),
        >>>     demo_site('DemoRegion_0004', site_poly4),
        >>>     demo_site('DemoRegion_0005', site_poly5),
        >>> ]
        >>> cropped_region, cropped_sites = crop_sites_to_region(region_gdf_crs84, sites)
        >>> cropped_sites = list(cropped_sites)
        >>> assert len(cropped_sites) == len(sites)
        >>> assert len(cropped_sites[0]['data']) == len(sites[0]['data'])
        >>> assert len(cropped_sites[1]['data']) == len(sites[1]['data'])
        >>> assert len(cropped_sites[2]['data']) == 0
        >>> assert len(cropped_region) == 2
    """
    from watch.utils import util_gis
    import pandas as pd
    output_crs = region_gdf_crs84.crs

    _row_type = region_gdf_crs84['type']
    region_rows_crs84 = region_gdf_crs84[_row_type == 'region']
    sitesum_rows_crs84 = region_gdf_crs84[_row_type == 'site_summary']

    num_sitesum_rows = len(sitesum_rows_crs84)
    num_region_rows = len(region_rows_crs84)
    num_total_rows = len(region_gdf_crs84)

    if num_region_rows != 1:
        raise AssertionError(f'can only have one region row with type region. Got {len(region_rows_crs84)}')

    if num_total_rows != num_sitesum_rows + num_region_rows:
        unique_types = region_gdf_crs84['type'].unique()
        raise AssertionError(f'Region file expected to only contain types of region and site_summary. Got {unique_types}')

    # Take only the first row, ignore site-summaries
    assert region_gdf_crs84.iloc[0].type == 'region'
    assert region_gdf_crs84.crs.name.startswith('WGS 84')

    site_summary_rows = region_gdf_crs84.iloc[1:]

    region_row_crs84 = region_rows_crs84.iloc[0:1]
    region_geom_crs84 = region_row_crs84.geometry.iloc[0]

    utm_epsg : int = util_gis.find_local_meter_epsg_crs(region_geom_crs84)
    region_row_utm = region_row_crs84.to_crs(utm_epsg)
    crop_geom_utm = region_row_utm.geometry.iloc[0]

    assert crop_geom_utm.is_valid

    # Crop the site summaries within the region file
    valid_site_summaries = crop_gdf_in_utm(
        site_summary_rows, crop_geom_utm, utm_epsg, output_crs)

    cropped_region = pd.concat([region_rows_crs84, valid_site_summaries])

    # Crop the site models to the region geometry
    cropped_sites = _cropper_gen(sites, crop_geom_utm, utm_epsg, output_crs)

    return cropped_region, cropped_sites


def _cropper_gen(sites, crop_geom_utm, utm_epsg, output_crs):
    import geopandas as gpd
    for site_info in sites:
        site_gdf_crs84: gpd.GeoDataFrame = site_info['data']
        valid_site_gdf_crs84 = crop_gdf_in_utm(
            site_gdf_crs84, crop_geom_utm, utm_epsg, output_crs)

        new_site_info: dict = site_info.copy()
        new_site_info['data'] = valid_site_gdf_crs84
        yield new_site_info


def crop_gdf_in_utm(gdf, crop_geom_utm, utm_epsg, output_crs):
    """
    Crop geometry in a geopandas data frame to specified bounds in UTM space.
    Filter out any rows where the cropped geometry is null or invalid.

    Args:
        gdf (geopandas.GeoDataFrame):
            The data to crop (in CRS84)

        crop_geom_utm (shapely.geometry.polygon.Polygon):
            The UTM polygon to crop to.

        utm_epsg (int):
            The UTM zone to work in

        output_crs (pyproj.crs.crs.CRS):
            The output CRS to wrap back into (should be CRS84)

    Returns:
        GeoDataFrame
    """
    gdf_utm = gdf.to_crs(utm_epsg)

    # Attempt to fix any polygon that became invalid after UTM projection
    invalid_proj = ~gdf_utm.geometry.is_valid
    if invalid_proj.any():
        gdf_utm.geometry[invalid_proj] = gdf_utm.geometry[invalid_proj].buffer(0)
    invalid_proj = ~gdf_utm.geometry.is_valid
    assert not invalid_proj.any()

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'invalid value', RuntimeWarning)
        warnings.filterwarnings('ignore', 'Self-intersection', RuntimeWarning)
        isect = gdf_utm.intersection(crop_geom_utm)
    flags = isect.area > 0

    valid_isect = isect[flags]
    valid_gdf_utm = gdf_utm[flags]
    valid_gdf_utm = valid_gdf_utm.assign(geometry=valid_isect)

    # Project back to the output CRS
    valid_gdf_crs84 = valid_gdf_utm.to_crs(output_crs)
    return valid_gdf_crs84


__config__ = CropSitesToRegionsConfig


if __name__ == '__main__':
    """
    CommandLine:
        DVC_DPATH=$(geowatch_dvc)
        python -m watch.cli.crop_sites_to_regions \
            --site_models=$DVC_DPATH/annotations/site_models/*KR*.geojson \
            --region_models=$DVC_DPATH/annotations/region_models/KR_R001.geojson \
            --new_site_dpath=$DVC_DPATH/tmp/new_site_models \
            --new_region_dpath=$DVC_DPATH/tmp/new_region_models
    """
    main(cmdline=True)



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union

import kwcoco
import torch.utils.data

from . import detector
from .datasets import S2Dataset, WVDataset

log = logging.getLogger(__name__)


class ModelInfo(ABC):

    @abstractmethod
    def create_dataset(self, coco_dset: Union[kwcoco.CocoDataset, str]) -> torch.utils.data.Dataset:
        """
        Create a torch dataset compatible with this model using a CocoDataset
        Args:
            coco_dset: CocoDataset of string filepath

        Returns: torch dataset

        """
        pass

    @property
    @abstractmethod
    def model_outputs(self):
        pass

    @abstractmethod
    def load_model(self, weights_filename: Path, device):
        pass


class S2ModelInfo(ModelInfo):
    """
    This model was trained on 13-band Sentinel-2 data with 5
    segmentation classes
    """

    def create_dataset(self, coco_dset):
        return S2Dataset(coco_dset)

    @property
    def model_outputs(self):
        return [
            'water', 'forest', 'field', 'impervious', 'barren',
        ]

    def load_model(self, weights_filename, device):
        assert len(self.model_outputs) == 5
        return detector.load_model(weights_filename,
                                   num_outputs=5,
                                   num_channels=13,
                                   device=device)


class WVModelInfo(ModelInfo):
    """
    This model was trained on 8-band WorldView-3 data with 5
    segmentation classes
    """

    def create_dataset(self, coco_dset):
        return WVDataset(coco_dset)

    @property
    def model_outputs(self):
        return [
            'water', 'forest', 'field', 'impervious', 'barren',
        ]

    def load_model(self, weights_filename, device):
        assert len(self.model_outputs) == 5
        return detector.load_model(weights_filename,
                                   num_outputs=5,
                                   num_channels=8,
                                   device=device)


__mapping = {
    'sentinel2': S2ModelInfo,
    'worldview': WVModelInfo,
}


def lookup_model_info(weights_filename: Path) -> ModelInfo:
    model_info_class = __mapping.get(weights_filename.stem)
    if not model_info_class:
        raise Exception('unknown weights file')
    return model_info_class()
